#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ------------------------------------------------------------------
# Filename: invsim.py
#  Purpose: Python Module for Instrument Correction (Seismology)
#   Author: Moritz Beyreuther, Yannik Behr
#    Email: moritz.beyreuther@geophysik.uni-muenchen.de
#
# Copyright (C) 2008-2012 Moritz Beyreuther, Yannik Behr
# --------------------------------------------------------------------
"""
Python Module for Instrument Correction (Seismology).
PAZ (Poles and zeros) information must be given in SEED convention, correction
to m/s.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
import ctypes as C  # NOQA
import math
import os
import warnings
from pathlib import Path

import numpy as np
import scipy.signal

from obspy.core.util.attribdict import AttribDict
from obspy.core.util.base import NamedTemporaryFile
from obspy.core.inventory.response import Response
from obspy.signal import util
from obspy.signal.detrend import simple as simple_detrend
from obspy.signal.headers import clibevresp
from obspy.signal.util import _npts2nfft


# Sensitivity is 2080 according to:
# Bormann, P. (ed.) (2002). IASPEI New Manual of Seismological Observatory
# Practice (NMSOP), GeoForschungsZentrum Potsdam, ISBN: 3-9808780-0-7,
# IASPEI Chapter 3, page 24
# (PITSA has 2800)
WOODANDERSON = {'poles': [-6.283 + 4.7124j, -6.283 - 4.7124j],
                'zeros': [0 + 0j], 'gain': 1.0, 'sensitivity': 2080}


def cosine_taper(npts, p=0.1, freqs=None, flimit=None, halfcosine=True,
                 sactaper=False):
    """
    Cosine Taper.

    :type npts: int
    :param npts: Number of points of cosine taper.
    :type p: float
    :param p: Decimal percentage of cosine taper (ranging from 0 to 1). Default
        is 0.1 (10%) which tapers 5% from the beginning and 5% form the end.
    :rtype: float NumPy :class:`~numpy.ndarray`
    :return: Cosine taper array/vector of length npts.
    :type freqs: NumPy :class:`~numpy.ndarray`
    :param freqs: Frequencies as, for example, returned by fftfreq
    :type flimit: list(float, float, float, float) or
        tuple(float, float, float, float)
    :param flimit: The list or tuple defines the four corner frequencies
        (f1, f2, f3, f4) of the cosine taper which is one between f2 and f3 and
        tapers to zero for f1 < f < f2 and f3 < f < f4.
    :type halfcosine: bool
    :param halfcosine: If True the taper is a half cosine function. If False it
        is a quarter cosine function.
    :type sactaper: bool
    :param sactaper: If set to True the cosine taper already tapers at the
        corner frequency (SAC behavior). By default, the taper has a value
        of 1.0 at the corner frequencies.

    .. rubric:: Example

    >>> tap = cosine_taper(100, 1.0)
    >>> tap2 = 0.5 * (1 + np.cos(np.linspace(np.pi, 2 * np.pi, 50)))
    >>> np.allclose(tap[0:50], tap2)
    True
    >>> npts = 100
    >>> p = 0.1
    >>> tap3 = cosine_taper(npts, p)
    >>> (tap3[int(npts*p/2):int(npts*(1-p/2))]==np.ones(int(npts*(1-p)))).all()
    True
    """
    if p < 0 or p > 1:
        msg = "Decimal taper percentage must be between 0 and 1."
        raise ValueError(msg)
    if p == 0.0 or p == 1.0:
        frac = int(npts * p / 2.0)
    else:
        frac = int(npts * p / 2.0 + 0.5)

    if freqs is not None and flimit is not None:
        fl1, fl2, fl3, fl4 = flimit
        idx1 = np.argmin(abs(freqs - fl1))
        idx2 = np.argmin(abs(freqs - fl2))
        idx3 = np.argmin(abs(freqs - fl3))
        idx4 = np.argmin(abs(freqs - fl4))
    else:
        idx1 = 0
        idx2 = frac - 1
        idx3 = npts - frac
        idx4 = npts - 1
    if sactaper:
        # in SAC the second and third
        # index are already tapered
        idx2 += 1
        idx3 -= 1

    # Very small data lengths or small decimal taper percentages can result in
    # idx1 == idx2 and idx3 == idx4. This breaks the following calculations.
    if idx1 == idx2:
        idx2 += 1
    if idx3 == idx4:
        idx3 -= 1

    # the taper at idx1 and idx4 equals zero and
    # at idx2 and idx3 equals one
    cos_win = np.zeros(npts)
    if halfcosine:
        # cos_win[idx1:idx2+1] =  0.5 * (1.0 + np.cos((np.pi * \
        #    (idx2 - np.arange(idx1, idx2+1)) / (idx2 - idx1))))
        cos_win[idx1:idx2 + 1] = 0.5 * (
            1.0 - np.cos((np.pi * (np.arange(idx1, idx2 + 1) - float(idx1)) /
                          (idx2 - idx1))))
        cos_win[idx2 + 1:idx3] = 1.0
        cos_win[idx3:idx4 + 1] = 0.5 * (
            1.0 + np.cos((np.pi * (float(idx3) - np.arange(idx3, idx4 + 1)) /
                          (idx4 - idx3))))
    else:
        cos_win[idx1:idx2 + 1] = np.cos(-(
            np.pi / 2.0 * (float(idx2) -
                           np.arange(idx1, idx2 + 1)) / (idx2 - idx1)))
        cos_win[idx2 + 1:idx3] = 1.0
        cos_win[idx3:idx4 + 1] = np.cos((
            np.pi / 2.0 * (float(idx3) -
                           np.arange(idx3, idx4 + 1)) / (idx4 - idx3)))

    # if indices are identical division by zero
    # causes NaN values in cos_win
    if idx1 == idx2:
        cos_win[idx1] = 0.0
    if idx3 == idx4:
        cos_win[idx3] = 0.0
    return cos_win


def cosine_sac_taper(freqs, flimit):
    """
    Create a cosine taper similar to SAC.

    Generate a cosine flank frequency domain taper similar to the one SAC
    applies before instrument response deconvolution. This acts as a bandpass
    filter when applied to the data in frequency space.

    :param freqs: frequency vector to use
    :type freqs: :class:`numpy.ndarray`
    :param flimit: sequence containing the 4 frequency limits
    :type flimit: tuple(float, float, float, float)
    :returns: taper
    :rtype: :class:`numpy.ndarray`

    The `flimit` parameter is a tuple of four frequency values `(f1, f2,
    f3, f4)`, the following plots illustrates the concept:

    .. plot::

        import matplotlib.pylab as plt
        import numpy as np
        from obspy.signal.invsim import cosine_sac_taper

        plt.figure(figsize=(10, 3))

        freqs = np.logspace(-2.01, 0, 2000)

        plt.vlines([0.015, 0.03, 0.2, 0.4], -0.1, 1.3, color="#89160F")
        plt.semilogx(freqs, cosine_sac_taper(freqs, (0.015, 0.03, 0.2, 0.4)),
                     lw=2, color="#4C72B0")

        props = {
            "bbox": dict(facecolor='white', edgecolor="0.5",
                 boxstyle="square,pad=0.2"),
            "va": "top", "ha": "center", "color": "#89160F",
            "size": "large"}
        plt.text(0.015, 1.25, "f1", **props)
        plt.text(0.03, 1.25, "f2", **props)
        plt.text(0.2, 1.25, "f3", **props)
        plt.text(0.4, 1.25, "f4", **props)

        plt.xlim(freqs[0], freqs[-1])
        plt.ylim(-0.1, 1.3)
        plt.ylabel("Taper Amplitude")
        plt.xlabel("Frequency [Hz]")
        plt.grid()
        plt.tight_layout()
        plt.show()
    """
    fl1, fl2, fl3, fl4 = flimit
    taper = np.zeros_like(freqs)

    a = (fl1 <= freqs) & (freqs <= fl2)
    taper[a] = 0.5 * (1.0 - np.cos(np.pi * (freqs[a] - fl1) / (fl2 - fl1)))

    b = (fl2 < freqs) & (freqs < fl3)
    taper[b] = 1.0

    c = (fl3 <= freqs) & (freqs <= fl4)
    taper[c] = 0.5 * (1.0 + np.cos(np.pi * (freqs[c] - fl3) / (fl4 - fl3)))

    return taper


def evalresp_for_frequencies(t_samp, frequencies, filename, date, station='*',
                             channel='*', network='*', locid='*', units="VEL",
                             debug=False):
    """
    Get the instrument response from a SEED RESP-file for select frequencies.

    Uses the evalresp library.

    :type t_samp: float
    :param t_samp: Sampling interval in seconds
    :type frequencies: list[float]
    :param frequencies: Discrete frequencies to calculate response for.
    :type filename: str or file or Path
    :param filename: SEED RESP-filename or open file like object with RESP
        information. Any object that provides a read() method will be
        considered to be a file like object.
    :type date: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param date: Date of interest
    :type station: str
    :param station: Station id
    :type channel: str
    :param channel: Channel id
    :type network: str
    :param network: Network id
    :type locid: str
    :param locid: Location id
    :type units: str
    :param units: Units to return response in. Can be either DIS, VEL or ACC
    :type debug: bool
    :param debug: Verbose output to stdout. Disabled by default.
    :rtype: :class:`numpy.ndarray` complex128
    :return: Frequency response from SEED RESP-file for given frequencies
    """
    if isinstance(filename, (str, Path)):
        with open(filename, 'rb') as fh:
            data = fh.read()
    elif hasattr(filename, 'read'):
        data = filename.read()
    # evalresp needs files with correct line separators depending on OS
    with NamedTemporaryFile() as fh:
        tempfile = fh.name
        fh.write(os.linesep.encode('ascii', 'strict').join(data.splitlines()))
        fh.close()

        # start at zero to get zero for offset/ DC of fft
        start_stage = C.c_int(-1)
        stop_stage = C.c_int(0)
        stdio_flag = C.c_int(0)
        sta = C.create_string_buffer(station.encode('ascii', 'strict'))
        cha = C.create_string_buffer(channel.encode('ascii', 'strict'))
        net = C.create_string_buffer(network.encode('ascii', 'strict'))
        locid = C.create_string_buffer(locid.encode('ascii', 'strict'))
        unts = C.create_string_buffer(units.encode('ascii', 'strict'))
        if debug:
            vbs = C.create_string_buffer(b"-v")
        else:
            vbs = C.create_string_buffer(b"")
        rtyp = C.create_string_buffer(b"CS")
        datime = C.create_string_buffer(
            date.format_seed().encode('ascii', 'strict'))
        fn = C.create_string_buffer(tempfile.encode('ascii', 'strict'))
        frequencies = np.asarray(frequencies)
        nfreqs = C.c_int(frequencies.shape[0])
        res = clibevresp.evresp(sta, cha, net, locid, datime, unts, fn,
                                frequencies, nfreqs, rtyp, vbs, start_stage,
                                stop_stage, stdio_flag, C.c_int(0))
        # optimizing performance, see
        # https://wiki.python.org/moin/PythonSpeed/PerformanceTips
        try:
            nfreqs, rfreqs, rvec = \
                res[0].nfreqs, res[0].freqs, res[0].rvec
        except ValueError:
            msg = "evalresp failed to calculate a response."
            raise ValueError(msg)
        h = np.empty(nfreqs, dtype=np.complex128)
        for i in range(nfreqs):
            h[i] = rvec[i].real + rvec[i].imag * 1j
        clibevresp.free_response(res)
        del nfreqs, rfreqs, rvec, res
    return h


def evalresp(t_samp, nfft, filename, date, station='*', channel='*',
             network='*', locid='*', units="VEL", freq=False,
             debug=False):
    """
    Get the instrument response from a SEED RESP-file.

    Uses the evalresp library.

    :type t_samp: float
    :param t_samp: Sampling interval in seconds
    :type nfft: int
    :param nfft: Number of FFT points of signal which needs correction
    :type filename: str or file
    :param filename: SEED RESP-filename or open file like object with RESP
        information. Any object that provides a read() method will be
        considered to be a file like object.
    :type date: :class:`~obspy.core.utcdatetime.UTCDateTime`
    :param date: Date of interest
    :type station: str
    :param station: Station id
    :type channel: str
    :param channel: Channel id
    :type network: str
    :param network: Network id
    :type locid: str
    :param locid: Location id
    :type units: str
    :param units: Units to return response in. Can be either DIS, VEL or ACC
    :type debug: bool
    :param debug: Verbose output to stdout. Disabled by default.
    :rtype: :class:`numpy.ndarray` complex128
    :return: Frequency response from SEED RESP-file of length nfft
    """
    fy = 1 / (t_samp * 2.0)
    # start at zero to get zero for offset/ DC of fft
    freqs = np.linspace(0, fy, nfft // 2 + 1)
    h = evalresp_for_frequencies(t_samp, freqs, filename, date, station,
                                 channel, network, locid, units, debug=debug)
    if freq:
        return h, freqs
    return h


def corn_freq_2_paz(fc, damp=0.707):
    """
    Convert corner frequency and damping coefficient to poles and zeros.

    2 zeros at position (0j, 0j) are given as output (m/s).

    :param fc: Corner frequency
    :type fc: float
    :param damp: Damping coefficient (default: 0.707)
    :type damp: float
    :return: Dictionary containing poles, zeros, gain and sensitivity
    :rtype: dict
    """
    poles = [-(damp + math.sqrt(1 - damp ** 2) * 1j) * 2 * np.pi * fc,
             -(damp - math.sqrt(1 - damp ** 2) * 1j) * 2 * np.pi * fc]
    return {'poles': poles, 'zeros': [0j, 0j], 'gain': 1, 'sensitivity': 1.0}


def paz_to_freq_resp(poles, zeros, scale_fac, t_samp, nfft, freq=False):
    """
    Convert Poles and Zeros (PAZ) to frequency response.

    The output contains the frequency zero which is the offset of the trace.

    :type poles: list[complex]
    :param poles: The poles of the transfer function
    :type zeros: list[complex]
    :param zeros: The zeros of the transfer function
    :type scale_fac: float
    :param scale_fac: Gain factor
    :type t_samp: float
    :param t_samp: Sampling interval in seconds
    :type nfft: int
    :param nfft: Number of FFT points of signal which needs correction
    :rtype: :class:`numpy.ndarray` complex128
    :return: Frequency response of PAZ of length nfft
    """
    n = nfft // 2
    b, a = scipy.signal.zpk2tf(zeros, poles, scale_fac)
    # a has to be a list for the scipy.signal.freqs() call later but zpk2tf()
    # strangely returns it as an integer.
    if not isinstance(a, np.ndarray) and a == 1.0:
        a = [1.0]
    fy = 1 / (t_samp * 2.0)
    # start at zero to get zero for offset / DC of fft
    f = np.linspace(0, fy, n + 1)
    _w, h = scipy.signal.freqs(b, a, f * 2 * np.pi)
    if freq:
        return h, f
    return h


def waterlevel(spec, wlev):
    """
    Get the absolute spectral value corresponding to dB wlev in spectrum spec.

    :param spec: The spectrum
    :param wlev: The water level
    """
    return np.abs(spec).max() * 10.0 ** (-wlev / 20.0)


def invert_spectrum(spec, wlev):
    """
    Invert Spectrum and shrink values under water-level of max spec amplitude.

    The water-level is given in db scale.

    :note: In place operations on spec, translated from PITSA spr_sinv.c
    :param spec: Spectrum as returned by :func:`numpy.fft.rfft`
    :param wlev: Water level to use
    """
    # Calculated water level in the scale of spec
    swamp = waterlevel(spec, wlev)

    # Find length in real fft frequency domain, spec is complex
    sqrt_len = np.abs(spec)
    # Set/scale length to swamp, but leave phase untouched
    # 0 sqrt_len will transform in np.nans when dividing by it
    idx = np.where((sqrt_len < swamp) & (sqrt_len > 0.0))
    spec[idx] *= swamp / sqrt_len[idx]
    found = len(idx[0])
    # Now invert the spectrum for values where sqrt_len is greater than
    # 0.0, see PITSA spr_sinv.c for details
    sqrt_len = np.abs(spec)  # Find length of new scaled spec
    inn = np.where(sqrt_len > 0.0)
    spec[inn] = 1.0 / spec[inn]
    # For numerical stability, set all zero length to zero, do not invert
    spec[sqrt_len == 0.0] = complex(0.0, 0.0)
    return found


def simulate_seismometer(
        data, samp_rate, paz_remove=None, paz_simulate=None,
        remove_sensitivity=True, simulate_sensitivity=True, water_level=600.0,
        zero_mean=True, taper=True, taper_fraction=0.05, pre_filt=None,
        seedresp=None, nfft_pow2=False, pitsasim=True, sacsim=False,
        shsim=False):
    """
    Simulate/Correct seismometer.

    :type data: NumPy :class:`~numpy.ndarray`
    :param data: Seismogram, detrend before hand (e.g. zero mean)
    :type samp_rate: float
    :param samp_rate: Sample Rate of Seismogram
    :type paz_remove: dict, None
    :param paz_remove: Dictionary containing keys 'poles', 'zeros', 'gain'
        (A0 normalization factor). poles and zeros must be a list of complex
        floating point numbers, gain must be of type float. Poles and Zeros are
        assumed to correct to m/s, SEED convention. Use None for no inverse
        filtering.
    :type paz_simulate: dict, None
    :param paz_simulate: Dictionary containing keys 'poles', 'zeros', 'gain'.
        Poles and zeros must be a list of complex floating point numbers, gain
        must be of type float. Or None for no simulation.
    :type remove_sensitivity: bool
    :param remove_sensitivity: Determines if data is divided by
        `paz_remove['sensitivity']` to correct for overall sensitivity of
        recording instrument (seismometer/digitizer) during instrument
        correction.
    :type simulate_sensitivity: bool
    :param simulate_sensitivity: Determines if data is multiplied with
        `paz_simulate['sensitivity']` to simulate overall sensitivity of
        new instrument (seismometer/digitizer) during instrument simulation.
    :type water_level: float
    :param water_level: Water_Level for spectrum to simulate
    :type zero_mean: bool
    :param zero_mean: If true the mean of the data is subtracted
    :type taper: bool
    :param taper: If true a cosine taper is applied.
    :type taper_fraction: float
    :param taper_fraction: Taper fraction of cosine taper to use
    :type pre_filt: list(float, float, float, float) or
        tuple(float, float, float, float)
    :param pre_filt: Apply a bandpass filter to the data trace before
        deconvolution. The list or tuple defines the four corner frequencies
        (f1,f2,f3,f4) of a cosine taper which is one between f2 and f3 and
        tapers to zero for f1 < f < f2 and f3 < f < f4.
    :type seedresp: dict, None
    :param seedresp: Dictionary contains keys 'filename', 'date', 'units'.
        'filename' is the path to a RESP-file generated from a dataless SEED
        volume (or a file like object with RESP information);
        'date' is a :class:`~obspy.core.utcdatetime.UTCDateTime` object for the
        date that the response function should be extracted for (can be omitted
        when calling simulate() on Trace/Stream. the Trace's starttime will
        then be used);
        'units' defines the units of the response function.
        Can be either 'DIS', 'VEL' or 'ACC'.
    :type nfft_pow2: bool
    :param nfft_pow2: Number of frequency points to use for FFT. If True,
        the exact power of two is taken (default in PITSA). If False the
        data are not zero-padded to the next power of two which makes a
        slower FFT but is then much faster for e.g. evalresp which scales
        with the FFT points.
    :type pitsasim: bool
    :param pitsasim: Choose parameters to match
        instrument correction as done by PITSA.
    :type sacsim: bool
    :param sacsim: Choose parameters to match
        instrument correction as done by SAC.
    :type shsim: bool
    :param shsim: Choose parameters to match
        instrument correction as done by Seismic Handler.
    :return: The corrected data are returned as :class:`numpy.ndarray` float64
        array. float64 is chosen to avoid numerical instabilities.

    This function works in the frequency domain, where nfft is the next power
    of len(data) to avoid wrap around effects during convolution. The inverse
    of the frequency response of the seismometer (``paz_remove``) is
    convolved with the spectrum of the data and with the frequency response
    of the seismometer to simulate (``paz_simulate``). A 5% cosine taper is
    taken before simulation. The data must be detrended (e.g.) zero mean
    beforehand. If paz_simulate=None only the instrument correction is done.
    In the latter case, a broadband filter can be applied to the data trace
    using pre_filt. This restricts the signal to the valid frequency band and
    thereby avoids artifacts due to amplification of frequencies outside of the
    instrument's passband (for a detailed discussion see
    *Of Poles and Zeros*, F. Scherbaum, Kluwer Academic Publishers).

    .. versionchanged:: 0.5.1
        The default for `remove_sensitivity` and `simulate_sensitivity` has
        been changed to ``True``. Old deprecated keyword arguments `paz`,
        `inst_sim`, `no_inverse_filtering` have been removed.
    """
    # Checking the types
    if not paz_remove and not paz_simulate and not seedresp:
        msg = "Neither inverse nor forward instrument simulation specified."
        raise TypeError(msg)

    for d in [paz_remove, paz_simulate]:
        if d is None:
            continue
        for key in ['poles', 'zeros', 'gain']:
            if key not in d:
                raise KeyError("Missing key: %s" % key)
    # Translated from PITSA: spr_resg.c
    delta = 1.0 / samp_rate
    #
    ndat = len(data)
    data = data.astype(np.float64)
    if zero_mean:
        data -= data.mean()
    if taper:
        if sacsim:
            data *= cosine_taper(ndat, taper_fraction,
                                 sactaper=sacsim, halfcosine=False)
        else:
            data *= cosine_taper(ndat, taper_fraction)
    # The number of points for the FFT has to be at least 2 * ndat (in
    # order to prohibit wrap around effects during convolution) cf.
    # Numerical Recipes p. 429 calculate next power of 2.
    if nfft_pow2:
        nfft = util.next_pow_2(2 * ndat)
    # evalresp scales directly with nfft, therefore taking the next power of
    # two has a greater negative performance impact than the slow down of a
    # not power of two in the FFT
    else:
        nfft = _npts2nfft(ndat)
    # Transform data in Fourier domain
    data = np.fft.rfft(data, n=nfft)
    # Inverse filtering = Instrument correction
    if paz_remove:
        freq_response, freqs = paz_to_freq_resp(
            paz_remove['poles'], paz_remove['zeros'], paz_remove['gain'],
            delta, nfft, freq=True)
    if seedresp:
        freq_response, freqs = evalresp(delta, nfft, seedresp['filename'],
                                        seedresp['date'],
                                        units=seedresp['units'], freq=True,
                                        network=seedresp['network'],
                                        station=seedresp['station'],
                                        locid=seedresp['location'],
                                        channel=seedresp['channel'])
        if not remove_sensitivity:
            msg = "remove_sensitivity is set to False, but since seedresp " + \
                  "is selected the overall sensitivity will be corrected " + \
                  " for anyway!"
            warnings.warn(msg)
    if paz_remove or seedresp:
        if pre_filt:
            # make cosine taper
            fl1, fl2, fl3, fl4 = pre_filt
            if sacsim:
                cos_win = cosine_sac_taper(freqs, flimit=(fl1, fl2, fl3, fl4))
            else:
                cos_win = cosine_taper(freqs.size, freqs=freqs,
                                       flimit=(fl1, fl2, fl3, fl4))
            data *= cos_win
        invert_spectrum(freq_response, water_level)
        data *= freq_response
        del freq_response
    # Forward filtering = Instrument simulation
    if paz_simulate:
        data *= paz_to_freq_resp(paz_simulate['poles'], paz_simulate['zeros'],
                                 paz_simulate['gain'], delta, nfft)

    data[-1] = abs(data[-1]) + 0.0j
    # transform data back into the time domain
    data = np.fft.irfft(data)[0:ndat]
    if pitsasim:
        # linear detrend
        data = simple_detrend(data)
    if shsim:
        # detrend using least squares
        data = scipy.signal.detrend(data, type="linear")
    # correct for involved overall sensitivities
    if paz_remove and remove_sensitivity and not seedresp:
        data /= paz_remove['sensitivity']
    if paz_simulate and simulate_sensitivity:
        data *= paz_simulate['sensitivity']
    return data


def paz_2_amplitude_value_of_freq_resp(paz, freq):
    """
    Returns Amplitude at one frequency for the given poles and zeros.

    :param paz: Given poles and zeros
    :param freq: Given frequency

    The amplitude of the freq is estimated according to "Of Poles and
    Zeros", Frank Scherbaum, p 43.

    .. rubric:: Example

    >>> paz = {'poles': [-4.44 + 4.44j, -4.44 - 4.44j],
    ...        'zeros': [0 + 0j, 0 + 0j],
    ...        'gain': 0.4}
    >>> amp = paz_2_amplitude_value_of_freq_resp(paz, 1)
    >>> print(round(amp, 7))
    0.2830262
    """
    jw = complex(0, 2 * np.pi * freq)  # angular frequency
    fac = complex(1, 0)
    for zero in paz['zeros']:  # numerator
        fac *= jw - zero
    for pole in paz['poles']:  # denominator
        fac /= jw - pole
    return abs(fac) * paz['gain']


def estimate_magnitude(paz, amplitude, timespan, h_dist):
    """
    Estimate local magnitude.

    Estimates local magnitude from poles and zeros or full response of given
    instrument, the peak to peak amplitude and the time span from peak to peak.
    Readings on two components can be used in magnitude estimation by providing
    lists for ``paz``, ``amplitude`` and ``timespan``.

    .. note::
        The input poles and zeros for the actual instrument in use need to
        convert input data units to velocity in ``m/s``. Metadata from
        accelerometers and other instruments might have to be adjusted
        accordingly. It is strongly recommended to supply
        :class:`~obspy.core.inventory.response.Response` objects instead of
        poles and zeros dictionaries, as these adjustment are then done
        automatically.

    :param paz: PAZ of the instrument [m/s] (as a dictionary) or response of
        the instrument (as :class:`~obspy.core.inventory.response.Response`) or
        list of the same
    :param amplitude: Peak to peak amplitude [counts] or list of the same
    :param timespan: Timespan of peak to peak amplitude [s] or list of the same
    :param h_dist: Hypocentral distance [km]
    :returns: Estimated local magnitude Ml

    .. note::
        Magnitude estimation according to Bakun & Joyner, 1984, Eq. (3) page
        1835. Bakun, W. H. and W. B. Joyner: The Ml scale in central
        California, Bull. Seismol. Soc. Am., 74, 1827-1843, 1984

    .. rubric:: Example

    >>> paz = {'poles': [-4.444+4.444j, -4.444-4.444j, -1.083+0j],
    ...        'zeros': [0+0j, 0+0j, 0+0j],
    ...        'gain': 1.0, 'sensitivity': 671140000.0}
    >>> mag = estimate_magnitude(paz, 3.34e6, 0.065, 0.255)
    >>> print(round(mag, 6))
    2.132873
    >>> mag = estimate_magnitude([paz, paz], [3.34e6, 5e6], [0.065, 0.1],
    ...                          0.255)
    >>> print(round(mag, 6))
    2.347618
    """
    # convert input to lists
    if not isinstance(paz, list) and not isinstance(paz, tuple):
        paz = [paz]
    # check if PAZ or Response objects are given and set correct functions to
    # calculate wood anderson amplitude(s)
    wood_anderson_amplitude_functions = []
    for i in paz:
        # unfortunately AttribDict is not a subclass of dict so we have to
        # explicitly include it here
        if isinstance(i, (dict, AttribDict)):
            wood_anderson_amplitude_functions.append(
                estimate_wood_anderson_amplitude)
        elif isinstance(i, Response):
            wood_anderson_amplitude_functions.append(
                estimate_wood_anderson_amplitude_using_response)
        else:
            msg = ("Unknown response specification (type '{}'). Use a "
                   "dictionary structure with poles and zeros information or "
                   "an obspy Response object.").format(type(i))
            raise TypeError(msg)
    if not isinstance(amplitude, list) and not isinstance(amplitude, tuple):
        amplitude = [amplitude]
    if not isinstance(timespan, list) and not isinstance(timespan, tuple):
        timespan = [timespan]
    # convert every input amplitude to Wood Anderson and calculate the mean
    wa_ampl_mean = 0.0
    count = 0
    for func, paz, amplitude, timespan in zip(
            wood_anderson_amplitude_functions, paz, amplitude, timespan):
        wa_ampl_mean += func(paz, amplitude, timespan)
        count += 1
    wa_ampl_mean /= count
    # mean of input amplitudes (if more than one) should be used in final
    # magnitude estimation (usually N and E components)
    magnitude = np.log10(wa_ampl_mean) + np.log10(h_dist / 100.0) + \
        0.00301 * (h_dist - 100.0) + 3.0
    return magnitude


def estimate_wood_anderson_amplitude(paz, amplitude, timespan):
    """
    Calculate the Wood-Anderson amplitude equivalent.

    Convert amplitude in counts measured of instrument with given Poles and
    Zeros information for use in :func:`estimate_magnitude`.
    Amplitude should be measured as full peak to peak amplitude, timespan as
    difference of the two readings.

    .. note::
        It is recommended to use
        :func:`estimate_wood_anderson_amplitude_using_response` with a
        :class:`~obspy.core.inventory.response.Response` object instead
        whenever possible as this prevents user errors with potentially
        misaligned in between units when using poles and zeros from e.g.
        accelerometers.

    :param paz: PAZ of the instrument [m/s] or list of the same
    :param amplitude: Peak to peak amplitude [counts] or list of the same
    :param timespan: Timespan of peak to peak amplitude [s] or list of the same
    :returns: Simulated zero to peak displacement amplitude on Wood Anderson
        seismometer [mm] for use in local magnitude estimation.
    """
    # analog to pitsa/plt/RCS/plt_wave.c,v, lines 4881-4891
    freq = 1.0 / (2 * timespan)
    wa_ampl = amplitude / 2.0  # half peak to peak amplitude
    wa_ampl /= (paz_2_amplitude_value_of_freq_resp(paz, freq) *
                paz['sensitivity'])
    wa_ampl *= paz_2_amplitude_value_of_freq_resp(WOODANDERSON, freq) * \
        WOODANDERSON['sensitivity']
    wa_ampl *= 1000  # convert to mm
    return wa_ampl


def estimate_wood_anderson_amplitude_using_response(response, amplitude,
                                                    timespan):
    """
    Estimate the Wood-Anderson amplitude with a given instrument response.

    Convert amplitude in counts measured of instrument with given response
    information for use in :func:`estimate_magnitude`.
    Amplitude should be measured as full peak to peak amplitude, timespan as
    difference of the two readings.

    :param response: response of the instrument
    :type response: :class:`obspy.core.inventory.response.Response`
    :param amplitude: Peak to peak amplitude [counts] or list of the same
    :type amplitude: float
    :param timespan: Timespan of peak to peak amplitude [s] or list of the same
    :type timespan: float
    :returns: Simulated zero to peak displacement amplitude on Wood Anderson
        seismometer [mm] for use in local magnitude estimation.
    """
    freq = 1.0 / (2 * timespan)
    wa_ampl = amplitude / 2.0  # half peak to peak amplitude
    response = response.get_evalresp_response_for_frequencies(
        [freq], output="VEL", start_stage=None, end_stage=None)[0]
    response_amplitude = np.absolute(response)
    wa_ampl /= response_amplitude
    wa_ampl *= paz_2_amplitude_value_of_freq_resp(WOODANDERSON, freq) * \
        WOODANDERSON['sensitivity']
    wa_ampl *= 1000  # convert to mm
    return wa_ampl


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
The gse2.core test suite.
"""
import copy
import warnings

import numpy as np

from obspy import Stream, Trace, UTCDateTime, read
from obspy.core.util import NamedTemporaryFile
from obspy.io.gse2.libgse2 import ChksumError
import pytest


class TestCore():
    """
    Test cases for libgse2 core interface
    """
    def test_read_via_obspy(self, testdata):
        """
        Read files via L{obspy.Trace}
        """
        gse2file = testdata['loc_RJOB20050831023349.z']
        testdata = [12, -10, 16, 33, 9, 26, 16, 7, 17, 6, 1, 3, -2]
        # read
        st = read(gse2file, verify_checksum=True)
        st.verify()
        tr = st[0]
        assert tr.stats['station'] == 'RJOB'
        assert tr.stats.npts == 12000
        assert tr.stats['sampling_rate'] == 200
        assert tr.stats.get('channel') == 'Z'
        assert round(abs(tr.stats.get('calib')-9.49e-02), 7) == 0
        assert tr.stats.gse2.get('vang') == -1.0
        assert tr.stats.gse2.get('hang') == -1.0
        assert tr.stats.gse2.get('calper') == 1.0
        assert tr.stats.gse2.get('instype') == ''
        assert round(abs(tr.stats.starttime.timestamp-1125455629.850), 6) == 0
        assert tr.data[0:13].tolist() == testdata

    def test_read_head_via_obspy(self, testdata):
        """
        Read header of files via L{obspy.Trace}
        """
        gse2file = testdata['loc_RJOB20050831023349.z']
        # read
        st = read(gse2file, headonly=True)
        tr = st[0]
        assert tr.stats['station'] == 'RJOB'
        assert tr.stats.npts == 12000
        assert tr.stats['sampling_rate'] == 200
        assert tr.stats.get('channel') == 'Z'
        assert round(abs(tr.stats.get('calib')-9.49e-02), 7) == 0
        assert tr.stats.gse2.get('vang') == -1.0
        assert tr.stats.gse2.get('calper') == 1.0
        assert round(abs(tr.stats.starttime.timestamp-1125455629.850), 6) == 0
        assert str(tr.data) == '[]'

    def test_read_and_write_via_obspy(self, testdata):
        """
        Read and Write files via L{obspy.Trace}
        """
        gse2file = testdata['loc_RNON20040609200559.z']
        # read trace
        st1 = read(gse2file)
        st1.verify()
        tr1 = st1[0]
        # write comparison trace
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            st2 = Stream()
            st2.traces.append(Trace())
            tr2 = st2[0]
            tr2.data = copy.deepcopy(tr1.data)
            tr2.stats = copy.deepcopy(tr1.stats)
            # raises "UserWarning: Bad value in GSE2 header field"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                st2.write(tempfile, format='GSE2')
            # read comparison trace
            st3 = read(tempfile)
        st3.verify()
        tr3 = st3[0]
        # check if equal
        assert tr3.stats['station'] == tr1.stats['station']
        assert tr3.stats.npts == tr1.stats.npts
        assert tr3.stats['sampling_rate'] == tr1.stats['sampling_rate']
        assert tr3.stats.get('channel') == tr1.stats.get('channel')
        assert tr3.stats.get('starttime') == tr1.stats.get('starttime')
        assert tr3.stats.get('calib') == tr1.stats.get('calib')
        assert tr3.stats.gse2.get('vang') == tr1.stats.gse2.get('vang')
        assert tr3.stats.gse2.get('calper') == tr1.stats.gse2.get('calper')
        np.testing.assert_equal(tr3.data, tr1.data)

    def test_read_and_write_streams_via_obspy(self, testdata):
        """
        Read and Write files containing multiple GSE2 parts via L{obspy.Trace}
        """
        files = [testdata['loc_RNON20040609200559.z'],
                 testdata['loc_RJOB20050831023349.z']]
        testdata = [12, -10, 16, 33, 9, 26, 16, 7, 17, 6, 1, 3, -2]
        # write test file containing multiple GSE2 parts
        with NamedTemporaryFile() as tf:
            for filename in files:
                with open(filename, 'rb') as f1:
                    # raises "UserWarning: Bad value in GSE2 header field"
                    with warnings.catch_warnings():
                        warnings.simplefilter('ignore', UserWarning)
                        tf.write(f1.read())
            tf.flush()
            st1 = read(tf.name)
        st1.verify()
        assert len(st1) == 2
        tr11 = st1[0]
        tr12 = st1[1]
        assert tr11.stats['station'] == 'RNON'
        assert tr12.stats['station'] == 'RJOB'
        assert tr12.data[0:13].tolist() == testdata
        # write and read
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            # raises "UserWarning: Bad value in GSE2 header field"
            with warnings.catch_warnings():
                warnings.simplefilter('ignore', UserWarning)
                st1.write(tmpfile, format='GSE2')
            st2 = read(tmpfile)
        st2.verify()
        assert len(st2) == 2
        tr21 = st1[0]
        tr22 = st1[1]
        assert tr21.stats['station'] == 'RNON'
        assert tr22.stats['station'] == 'RJOB'
        assert tr22.data[0:13].tolist() == testdata
        np.testing.assert_equal(tr21.data, tr11.data)
        np.testing.assert_equal(tr22.data, tr12.data)

    def test_write_integers_via_obspy(self):
        """
        Write file test via L{obspy.Trace}.
        """
        npts = 1000
        # data cloud of integers - float won't work!
        np.random.seed(815)  # make test reproducible
        data = np.random.randint(-1000, 1000, npts).astype(np.int32)
        stats = {'network': 'BW', 'station': 'TEST', 'location': '',
                 'channel': 'EHE', 'npts': npts, 'sampling_rate': 200.0}
        start = UTCDateTime(2000, 1, 1)
        stats['starttime'] = start
        tr = Trace(data=data, header=stats)
        st = Stream([tr])
        st.verify()
        # write
        with NamedTemporaryFile() as tf:
            tempfile = tf.name
            st.write(tempfile, format="GSE2")
            # read again
            stream = read(tempfile)
        stream.verify()
        np.testing.assert_equal(data, stream[0].data)
        # test default attributes
        assert 'CM6' == stream[0].stats.gse2.datatype
        assert -1 == stream[0].stats.gse2.vang
        assert 1.0 == stream[0].stats.gse2.calper
        assert 1.0 == stream[0].stats.calib

    def test_tab_complete_stats(self, testdata):
        """
        Read files via L{obspy.Trace}
        """
        gse2file = testdata['loc_RJOB20050831023349.z']
        # read
        tr = read(gse2file)[0]
        assert 'station' in dir(tr.stats)
        assert 'npts' in dir(tr.stats)
        assert 'sampling_rate' in dir(tr.stats)
        assert tr.stats['station'] == 'RJOB'
        assert tr.stats.npts == 12000
        assert tr.stats['sampling_rate'] == 200

    def test_write_wrong_format(self):
        """
        Write floating point encoded data
        """
        np.random.seed(815)
        st = Stream([Trace(data=np.random.randn(1000))])
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            with pytest.raises(Exception):
                st.write(tmpfile, format="GSE2")

    def test_read_with_wrong_checksum(self, testdata):
        """
        Test if additional kwarg verify_chksum can be given
        """
        # read original file
        gse2file = testdata['loc_RJOB20050831023349.z.wrong_chksum']
        # should not fail
        read(gse2file, verify_chksum=False)
        # should fail
        with pytest.raises(ChksumError):
            read(gse2file, verify_chksum=True)

    def test_read_with_wrong_parameters(self, testdata):
        """
        Test if additional kwargs can be given
        """
        # read original file
        gse2file = testdata['loc_RJOB20050831023349.z.wrong_chksum']
        # add wrong starttime flag of mseed, should also not fail
        read(gse2file, verify_chksum=False, starttime=None)

    def test_read_gse1_via_obspy(self, testdata):
        """
        Read files via L{obspy.Trace}
        """
        gse1file = testdata['loc_STAU20031119011659.z']
        testdata = [-818, -814, -798, -844, -806, -818, -800, -790, -818, -780]
        # read
        st = read(gse1file, verify_checksum=True)
        st.verify()
        tr = st[0]
        assert tr.stats['station'] == 'LE0083'
        assert tr.stats.npts == 3000
        assert round(abs(tr.stats['sampling_rate']-124.9999924), 7) == 0
        assert tr.stats.get('channel') == '  Z'
        assert round(abs(tr.stats.get('calib')-16.0000001), 7) == 0
        assert str(tr.stats.starttime) == '2003-11-19T01:16:59.990000Z'
        assert tr.data[0:10].tolist() == testdata

    def test_read_gse1_head_via_obspy(self, testdata):
        """
        Read header via L{obspy.Trace}
        """
        gse1file = testdata['loc_STAU20031119011659.z']
        # read
        st = read(gse1file, headonly=True)
        tr = st[0]
        assert tr.stats['station'] == 'LE0083'
        assert tr.stats.npts == 3000
        assert round(abs(tr.stats['sampling_rate']-124.9999924), 7) == 0
        assert tr.stats.get('channel') == '  Z'
        assert round(abs(tr.stats.get('calib')-16.0000001), 7) == 0
        assert str(tr.stats.starttime) == '2003-11-19T01:16:59.990000Z'

    def test_read_intv_gse1_via_obspy(self, testdata):
        """
        Read file via L{obspy.Trace}
        """
        gse1file = testdata['GRF_031102_0225.GSE.wrong_chksum']
        data1 = [-334, -302, -291, -286, -266, -252, -240, -214]
        data2 = [-468, -480, -458, -481, -481, -435, -432, -389]

        # verify checksum fails
        with pytest.raises(ChksumError):
            read(gse1file, verify_chksum=True)
        # reading header only
        st = read(gse1file, headonly=True)
        assert len(st) == 2
        # reading without checksum verification
        st = read(gse1file, verify_chksum=False)
        st.verify()
        # first trace
        assert len(st) == 2
        assert st[0].stats['station'] == 'GRA1'
        assert st[0].stats.npts == 6000
        assert round(abs(st[0].stats['sampling_rate']-19.9999997), 7) == 0
        assert st[0].stats.get('channel') == ' BZ'
        assert round(abs(st[0].stats.get('calib')-0.9900001), 7) == 0
        assert st[0].stats.starttime == \
            UTCDateTime('2003-11-02T02:25:00.000000Z')
        # second trace
        assert len(st) == 2
        assert st[1].stats['station'] == 'GRA1'
        assert st[1].stats.npts == 6000
        assert round(abs(st[1].stats['sampling_rate']-19.9999997), 7) == 0
        assert st[1].stats.get('channel') == ' BN'
        assert round(abs(st[1].stats.get('calib')-0.9200001), 7) == 0
        assert st[1].stats.starttime == \
            UTCDateTime('2003-11-02T02:25:00.000000Z')
        # check first 8 samples
        assert st[0].data[0:8].tolist() == data1
        # check last 8 samples
        assert st[1].data[-8:].tolist() == data2

    def test_read_dos(self, testdata):
        """
        Read file with dos newlines / encoding, that is
        Line Feed (LF) and Carriage Return (CR)
        see #355
        """
        filedos = testdata['loc_RJOB20050831023349_first100_dos.z']
        fileunix = testdata['loc_RJOB20050831023349.z']
        st = read(filedos, verify_chksum=True)
        st2 = read(fileunix, verify_chksum=True)
        np.testing.assert_equal(st[0].data, st2[0].data[:100])
        assert st[0].stats['station'] == 'RJOB'

    def test_read_apply_calib(self, testdata):
        """
        Tests apply_calib parameter in read method.
        """
        gse2file = testdata['loc_RJOB20050831023349.z']
        testdata = [12, -10, 16, 33, 9, 26, 16, 7, 17, 6, 1, 3, -2]
        # read w/ apply_calib = False
        st = read(gse2file, apply_calib=False)
        tr = st[0]
        assert tr.data[0:13].tolist() == testdata
        # read w/ apply_calib = True
        st = read(gse2file, apply_calib=True)
        tr = st[0]
        testdata = [n * tr.stats.calib for n in testdata]
        assert tr.data[0:13].tolist() == testdata

    def test_write_and_read_correct_network(self):
        """
        Tests that writing and reading the STA2 line works (otherwise the
        network code of the data is missing), even if some details like e.g.
        latitude are not present.
        """
        tr = Trace(np.arange(5, dtype=np.int32))
        tr.stats.network = "BW"
        with NamedTemporaryFile() as tf:
            tmpfile = tf.name
            tr.write(tmpfile, format='GSE2')
            tr = read(tmpfile)[0]
        assert tr.stats.network == "BW"

    def test_read_gse2_int_datatype(self, testdata):
        """
        Test reading of GSE2 files with data type INT.
        """
        gse2file = testdata['boa___00_07a.gse']
        testdata = [-4, -4, 1, 3, 2, -3, -6, -4, 2, 5]
        # read
        st = read(gse2file, verify_checksum=True)
        st.verify()
        tr = st[0]
        assert tr.stats['station'] == 'BBOA'
        assert tr.stats.npts == 6784
        assert round(abs(tr.stats['sampling_rate']-50.0), 7) == 0
        assert tr.stats.get('channel') == 'CPZ'
        assert round(abs(tr.stats.get('calib')-0.313), 7) == 0
        assert str(tr.stats.starttime) == '1990-04-07T00:07:33.000000Z'
        assert tr.data[0:10].tolist() == testdata



# -*- coding: utf-8 -*-
"""
The obspy.signal.trigger test suite.
"""
import gzip
import re
import warnings
from ctypes import ArgumentError

import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal, assert_array_equal

from obspy import Stream, UTCDateTime, read
from obspy.signal.trigger import (
    ar_pick, classic_sta_lta, classic_sta_lta_py, coincidence_trigger, pk_baer,
    recursive_sta_lta, recursive_sta_lta_py, trigger_onset, aic_simple,
    energy_ratio, modified_energy_ratio)
from obspy.signal.util import clibsignal


def aic_simple_python(a):
    if len(a) <= 2:
        return np.zeros(len(a), dtype=np.float64)
    a = np.asarray(a)
    aic_cf = np.zeros(a.size - 1, dtype=np.float64)
    with np.errstate(divide='ignore'):
        aic_cf[0] = (a.size - 2) * np.log(np.var(a[1:]))
        aic_cf[-1] = (a.size - 1) * np.log(np.var(a[:-1]))
        for ii in range(2, a.size - 1):
            var1 = np.log(np.var(a[:ii]))
            var2 = np.log(np.var(a[ii:]))
            val1 = ii * var1
            val2 = (a.size - ii - 1) * var2
            aic_cf[ii - 1] = (val1 + val2)
    aic_cf = np.r_[aic_cf, aic_cf[-1]]
    return aic_cf


class TestTrigger():
    """
    Test cases for obspy.trigger
    """
    @classmethod
    def setup_class(cls):
        # random seed data
        np.random.seed(815)
        cls.data = np.random.randn(int(1e5))

    def test_rec_sta_lta_c(self):
        """
        Test case for ctypes version of recursive_sta_lta
        """
        nsta, nlta = 5, 10
        c1 = recursive_sta_lta(self.data, nsta, nlta)
        assert round(abs(c1[99]-0.80810165), 7) == 0
        assert round(abs(c1[100]-0.75939449), 7) == 0
        assert round(abs(c1[101]-0.91763978), 7) == 0
        assert round(abs(c1[102]-0.97465004), 7) == 0

    def test_rec_sta_lta_python(self):
        """
        Test case for python version of recursive_sta_lta
        """
        nsta, nlta = 5, 10
        c2 = recursive_sta_lta_py(self.data, nsta, nlta)
        assert round(abs(c2[99]-0.80810165), 7) == 0
        assert round(abs(c2[100]-0.75939449), 7) == 0
        assert round(abs(c2[101]-0.91763978), 7) == 0
        assert round(abs(c2[102]-0.97465004), 7) == 0

    def test_rec_sta_lta_raise(self):
        """
        Type checking recursive_sta_lta
        """
        ndat = 1
        charfct = np.empty(ndat, dtype=np.float64)
        with pytest.raises(ArgumentError):
            clibsignal.recstalta(
                [1], charfct, ndat, 5, 10)
        with pytest.raises(ArgumentError):
            clibsignal.recstalta(
                np.array([1], dtype=np.int32), charfct, ndat, 5, 10)

    def test_pk_baer(self, testdata):
        """
        Test pk_baer against implementation for UNESCO short course
        """
        with gzip.open(testdata['manz_waldk.a01.gz']) as f:
            data = np.loadtxt(f, dtype=np.float32)
        df, ntdownmax, ntupevent, thr1, thr2, npreset_len, np_dur = \
            (200.0, 20, 60, 7.0, 12.0, 100, 100)
        nptime, pfm = pk_baer(data, df, ntdownmax, ntupevent,
                              thr1, thr2, npreset_len, np_dur)
        assert nptime == 17545
        assert pfm == 'IPU0'

    def test_pk_baer_cf(self, testdata):
        """
        Test pk_baer against implementation for UNESCO short course
        """
        with gzip.open(testdata['manz_waldk.a01.gz']) as f:
            data = np.loadtxt(f, dtype=np.float32)
        df, ntdownmax, ntupevent, thr1, thr2, npreset_len, np_dur = \
            (200.0, 20, 60, 7.0, 12.0, 100, 100)
        nptime, pfm, cf = pk_baer(data, df, ntdownmax, ntupevent,
                                  thr1, thr2, npreset_len, np_dur,
                                  return_cf=True)
        assert nptime == 17545
        assert pfm == 'IPU0'
        assert len(cf) == 119999

    def test_aic_simple_constant_data(self):
        data = [1] * 10
        # all negative inf
        assert_array_equal(aic_simple(data), -np.inf)

    def test_aic_simple_small_size(self):
        data = [3, 4]
        assert_array_equal(aic_simple(data), [0, 0])

    def test_aic_simple(self):
        np.random.seed(0)
        data = np.random.rand(100)
        aic = aic_simple(data)
        assert len(aic) == len(data)
        aic_true = aic_simple_python(data)
        assert_array_almost_equal(aic, aic_true)

    def test_ar_pick(self, testdata):
        """
        Test ar_pick against implementation for UNESCO short course
        """
        data = []
        for channel in ['z', 'n', 'e']:
            filename = 'loc_RJOB20050801145719850.' + channel
            data.append(np.loadtxt(testdata[filename], dtype=np.float32))
        # some default arguments
        samp_rate, f1, f2, lta_p, sta_p, lta_s, sta_s, m_p, m_s, l_p, l_s = \
            200.0, 1.0, 20.0, 1.0, 0.1, 4.0, 1.0, 2, 8, 0.1, 0.2
        ptime, stime = ar_pick(data[0], data[1], data[2], samp_rate, f1, f2,
                               lta_p, sta_p, lta_s, sta_s, m_p, m_s, l_p, l_s)
        assert round(abs(ptime-30.6350002289), 7) == 0
        # seems to be strongly machine dependent, go for int for 64 bit
        # self.assertEqual(int(stime + 0.5), 31)
        assert abs(stime-31.165) < 0.05

        # All three arrays must have the same length, otherwise an error is
        # raised.
        msg = "All three data arrays must have the same length."
        with pytest.raises(ValueError, match=msg):
            ar_pick(data[0], data[1], np.zeros(1), samp_rate, f1, f2, lta_p,
                    sta_p, lta_s, sta_s, m_p, m_s, l_p, l_s)

    def test_ar_pick_low_amplitude(self, testdata):
        """
        Test ar_pick with low amplitude data
        """
        data = []
        for channel in ['z', 'n', 'e']:
            filename = 'loc_RJOB20050801145719850.' + channel
            data.append(np.loadtxt(testdata[filename], dtype=np.float32))

        # articially reduce signal amplitude
        for d in data:
            d /= 10.0 * d.max()

        # some default arguments
        samp_rate, f1, f2, lta_p, sta_p, lta_s, sta_s, m_p, m_s, l_p, l_s = \
            200.0, 1.0, 20.0, 1.0, 0.1, 4.0, 1.0, 2, 8, 0.1, 0.2
        ptime, stime = ar_pick(data[0], data[1], data[2], samp_rate, f1, f2,
                               lta_p, sta_p, lta_s, sta_s, m_p, m_s, l_p, l_s)
        assert round(abs(ptime-30.6350002289), 7) == 0
        # seems to be strongly machine dependent, go for int for 64 bit
        # self.assertAlmostEqual(stime, 31.2800006866)
        assert int(stime + 0.5) == 31

    def test_ar_pick_3552(self, testdata):
        """
        Regression test for PR 3552
        """
        tr = read(testdata['arpick_pr3552.mseed'], 'MSEED')[0]
        # no need to test any values etc, the following was crashing
        # obspy/python/pytest with a segfault
        ar_pick(tr.data, np.arange(len(tr)), np.arange(len(tr)),
                tr.stats.sampling_rate, 1.0, 20.0, 1.0, 0.1, 4.0, 1.0, 2, 8,
                0.1, 0.2)

    def test_trigger_onset(self):
        """
        Test trigger onset function
        """
        on_of = np.array([[6.0, 31], [69, 94], [131, 181], [215, 265],
                          [278, 315], [480, 505], [543, 568], [605, 631]])
        cft = np.concatenate((np.sin(np.arange(0, 5 * np.pi, 0.1)) + 1,
                              np.sin(np.arange(0, 5 * np.pi, 0.1)) + 2.1,
                              np.sin(np.arange(0, 5 * np.pi, 0.1)) + 0.4,
                              np.sin(np.arange(0, 5 * np.pi, 0.1)) + 1))
        picks = trigger_onset(cft, 1.5, 1.0, max_len=50)
        np.testing.assert_array_equal(picks, on_of)
        # check that max_len_delete drops the picks
        picks_del = trigger_onset(cft, 1.5, 1.0, max_len=50,
                                  max_len_delete=True)
        np.testing.assert_array_equal(
            picks_del, on_of[np.array([0, 1, 5, 6, 7])])
        #
        # set True for visual understanding the tests
        if False:  # pragma: no cover
            import matplotlib.pyplot as plt
            plt.plot(cft)
            plt.hlines([1.5, 1.0], 0, len(cft))
            on_of = np.array(on_of)
            plt.vlines(picks[:, 0], 1.0, 2.0, color='g', linewidth=2,
                       label="ON max_len")
            plt.vlines(picks[:, 1], 0.5, 1.5, color='r', linewidth=2,
                       label="OF max_len")
            plt.vlines(picks_del[:, 0] + 2, 1.0, 2.0, color='y', linewidth=2,
                       label="ON max_len_delete")
            plt.vlines(picks_del[:, 1] + 2, 0.5, 1.5, color='b', linewidth=2,
                       label="OF max_len_delete")
            plt.legend()
            plt.show()

    def test_trigger_onset_issue_2891(self, testdata):
        """
        Regression test for issue 2891

        This used to raise an error if a trigger was activated near the end of
        the trace, and all sample values after that trigger on threshold are
        above the designated off threshold. So basically this can only happen
        if the on threshold is below the off threshold, which is kind of
        unusual, but we fixed it nevertheless, since people can run into this
        playing around with different threshold settings
        """
        tr = read(testdata['BW.UH1._.EHZ.D.2010.147.a.slist.gz'])[0]
        cft = recursive_sta_lta(tr.data, 5, 30)
        trigger_onset(cft, 2.5, 3.2)

    def test_coincidence_trigger(self, testdata):
        """
        Test network coincidence trigger.
        """
        st = Stream()
        files = ["BW.UH1._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH2._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH3._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH4._.EHZ.D.2010.147.cut.slist.gz"]
        for filename in files:
            st += read(testdata[filename])
        # some prefiltering used for UH network
        st.filter('bandpass', freqmin=10, freqmax=20)

        # 1. no weighting, no stations specified, good settings
        # => 3 events, no false triggers
        # for the first test we make some additional tests regarding types
        res = coincidence_trigger("recstalta", 3.5, 1, st.copy(), 3, sta=0.5,
                                  lta=10)
        assert isinstance(res, list)
        assert len(res) == 3
        expected_keys = ['time', 'coincidence_sum', 'duration', 'stations',
                         'trace_ids']
        expected_types = [UTCDateTime, float, float, list, list]
        for item in res:
            assert isinstance(item, dict)
            for key, _type in zip(expected_keys, expected_types):
                assert key in item
                assert isinstance(item[key], _type)
        assert res[0]['time'] > UTCDateTime("2010-05-27T16:24:31")
        assert res[0]['time'] < UTCDateTime("2010-05-27T16:24:35")
        assert 4.2 < res[0]['duration'] < 4.8
        assert res[0]['stations'] == ['UH3', 'UH2', 'UH1', 'UH4']
        assert res[0]['coincidence_sum'] == 4
        assert res[1]['time'] > UTCDateTime("2010-05-27T16:26:59")
        assert res[1]['time'] < UTCDateTime("2010-05-27T16:27:03")
        assert 3.2 < res[1]['duration'] < 3.7
        assert res[1]['stations'] == ['UH2', 'UH3', 'UH1']
        assert res[1]['coincidence_sum'] == 3
        assert res[2]['time'] > UTCDateTime("2010-05-27T16:27:27")
        assert res[2]['time'] < UTCDateTime("2010-05-27T16:27:33")
        assert 4.2 < res[2]['duration'] < 4.4
        assert res[2]['stations'] == ['UH3', 'UH2', 'UH1', 'UH4']
        assert res[2]['coincidence_sum'] == 4

        # 2. no weighting, station selection
        # => 2 events, no false triggers
        trace_ids = ['BW.UH1..SHZ', 'BW.UH3..SHZ', 'BW.UH4..EHZ']
        # raises "UserWarning: At least one trace's ID was not found"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            re = coincidence_trigger("recstalta", 3.5, 1, st.copy(), 3,
                                     trace_ids=trace_ids, sta=0.5, lta=10)
            assert len(w) == 1
            assert "At least one trace's ID was not" in str(w[0])
        assert len(re) == 2
        assert re[0]['time'] > UTCDateTime("2010-05-27T16:24:31")
        assert re[0]['time'] < UTCDateTime("2010-05-27T16:24:35")
        assert 4.2 < re[0]['duration'] < 4.8
        assert re[0]['stations'] == ['UH3', 'UH1', 'UH4']
        assert re[0]['coincidence_sum'] == 3
        assert re[1]['time'] > UTCDateTime("2010-05-27T16:27:27")
        assert re[1]['time'] < UTCDateTime("2010-05-27T16:27:33")
        assert 4.2 < re[1]['duration'] < 4.4
        assert re[1]['stations'] == ['UH3', 'UH1', 'UH4']
        assert re[1]['coincidence_sum'] == 3

        # 3. weighting, station selection
        # => 3 events, no false triggers
        trace_ids = {'BW.UH1..SHZ': 0.4, 'BW.UH2..SHZ': 0.35,
                     'BW.UH3..SHZ': 0.4, 'BW.UH4..EHZ': 0.25}
        res = coincidence_trigger("recstalta", 3.5, 1, st.copy(), 1.0,
                                  trace_ids=trace_ids, sta=0.5, lta=10)
        assert len(res) == 3
        assert res[0]['time'] > UTCDateTime("2010-05-27T16:24:31")
        assert res[0]['time'] < UTCDateTime("2010-05-27T16:24:35")
        assert 4.2 < res[0]['duration'] < 4.8
        assert res[0]['stations'] == ['UH3', 'UH2', 'UH1', 'UH4']
        assert res[0]['coincidence_sum'] == 1.4
        assert res[1]['time'] > UTCDateTime("2010-05-27T16:26:59")
        assert res[1]['time'] < UTCDateTime("2010-05-27T16:27:03")
        assert 3.2 < res[1]['duration'] < 3.7
        assert res[1]['stations'] == ['UH2', 'UH3', 'UH1']
        assert res[1]['coincidence_sum'] == 1.15
        assert res[2]['time'] > UTCDateTime("2010-05-27T16:27:27")
        assert res[2]['time'] < UTCDateTime("2010-05-27T16:27:33")
        assert 4.2 < res[2]['duration'] < 4.4
        assert res[2]['stations'] == ['UH3', 'UH2', 'UH1', 'UH4']
        assert res[2]['coincidence_sum'] == 1.4

        # 4. weighting, station selection, max_len
        # => 2 events, no false triggers, small event does not overlap anymore
        trace_ids = {'BW.UH1..SHZ': 0.6, 'BW.UH2..SHZ': 0.6}
        # raises "UserWarning: At least one trace's ID was not found"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            re = coincidence_trigger("recstalta", 3.5, 1, st.copy(), 1.2,
                                     trace_ids=trace_ids,
                                     max_trigger_length=0.13, sta=0.5, lta=10)
            assert len(w) == 2
            assert "At least one trace's ID was not" in str(w[0])
            assert "At least one trace's ID was not" in str(w[1])
        assert len(re) == 2
        assert re[0]['time'] > UTCDateTime("2010-05-27T16:24:31")
        assert re[0]['time'] < UTCDateTime("2010-05-27T16:24:35")
        assert 0.2 < re[0]['duration'] < 0.3
        assert re[0]['stations'] == ['UH2', 'UH1']
        assert re[0]['coincidence_sum'] == 1.2
        assert re[1]['time'] > UTCDateTime("2010-05-27T16:27:27")
        assert re[1]['time'] < UTCDateTime("2010-05-27T16:27:33")
        assert 0.18 < re[1]['duration'] < 0.2
        assert re[1]['stations'] == ['UH2', 'UH1']
        assert re[1]['coincidence_sum'] == 1.2

        # 5. station selection, extremely sensitive settings
        # => 4 events, 1 false triggers
        # raises "UserWarning: At least one trace's ID was not found"
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            res = coincidence_trigger("recstalta", 2.5, 1, st.copy(), 2,
                                      trace_ids=['BW.UH1..SHZ', 'BW.UH3..SHZ'],
                                      sta=0.3, lta=5)
            assert len(w) == 2
            assert "At least one trace's ID was not" in str(w[0])
            assert "At least one trace's ID was not" in str(w[1])
        assert len(res) == 5
        assert res[3]['time'] > UTCDateTime("2010-05-27T16:27:01")
        assert res[3]['time'] < UTCDateTime("2010-05-27T16:27:02")
        assert 1.5 < res[3]['duration'] < 1.7
        assert res[3]['stations'] == ['UH3', 'UH1']
        assert res[3]['coincidence_sum'] == 2.0

        # 6. same as 5, gappy stream
        # => same as 5 (almost, duration of 1 event changes by 0.02s)
        st2 = st.copy()
        tr1 = st2.pop(0)
        t1 = tr1.stats.starttime
        t2 = tr1.stats.endtime
        td = t2 - t1
        tr1a = tr1.slice(starttime=t1, endtime=t1 + 0.45 * td)
        tr1b = tr1.slice(starttime=t1 + 0.6 * td, endtime=t1 + 0.94 * td)
        st2.insert(1, tr1a)
        st2.insert(3, tr1b)
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter('always', UserWarning)
            res = coincidence_trigger("recstalta", 2.5, 1, st2, 2,
                                      trace_ids=['BW.UH1..SHZ', 'BW.UH3..SHZ'],
                                      sta=0.3, lta=5)
            assert len(w) == 2
            assert "At least one trace's ID was not" in str(w[0])
            assert "At least one trace's ID was not" in str(w[1])
        assert len(res) == 5
        assert res[3]['time'] > UTCDateTime("2010-05-27T16:27:01")
        assert res[3]['time'] < UTCDateTime("2010-05-27T16:27:02")
        assert 1.5 < res[3]['duration'] < 1.7
        assert res[3]['stations'] == ['UH3', 'UH1']
        assert res[3]['coincidence_sum'] == 2.0

        # 7. same as 3 but modify input trace ids and check output of trace_ids
        # and other additional information with ``details=True``
        st2 = st.copy()
        st2[0].stats.network = "XX"
        st2[1].stats.location = "99"
        st2[1].stats.network = ""
        st2[1].stats.location = "99"
        st2[1].stats.channel = ""
        st2[2].stats.channel = "EHN"
        st2[3].stats.network = ""
        st2[3].stats.channel = ""
        st2[3].stats.station = ""
        trace_ids = {'XX.UH1..SHZ': 0.4, '.UH2.99.': 0.35,
                     'BW.UH3..EHN': 0.4, '...': 0.25}
        res = coincidence_trigger("recstalta", 3.5, 1, st2, 1.0,
                                  trace_ids=trace_ids, details=True,
                                  sta=0.5, lta=10)
        assert len(res) == 3
        assert res[0]['time'] > UTCDateTime("2010-05-27T16:24:31")
        assert res[0]['time'] < UTCDateTime("2010-05-27T16:24:35")
        assert 4.2 < res[0]['duration'] < 4.8
        assert res[0]['stations'] == ['UH3', 'UH2', 'UH1', '']
        assert res[0]['trace_ids'][0] == st2[2].id
        assert res[0]['trace_ids'][1] == st2[1].id
        assert res[0]['trace_ids'][2] == st2[0].id
        assert res[0]['trace_ids'][3] == st2[3].id
        assert res[0]['coincidence_sum'] == 1.4
        assert res[1]['time'] > UTCDateTime("2010-05-27T16:26:59")
        assert res[1]['time'] < UTCDateTime("2010-05-27T16:27:03")
        assert 3.2 < res[1]['duration'] < 3.7
        assert res[1]['stations'] == ['UH2', 'UH3', 'UH1']
        assert res[1]['trace_ids'][0] == st2[1].id
        assert res[1]['trace_ids'][1] == st2[2].id
        assert res[1]['trace_ids'][2] == st2[0].id
        assert res[1]['coincidence_sum'] == 1.15
        assert res[2]['time'] > UTCDateTime("2010-05-27T16:27:27")
        assert res[2]['time'] < UTCDateTime("2010-05-27T16:27:33")
        assert 4.2 < res[2]['duration'] < 4.4
        assert res[2]['stations'] == ['UH3', 'UH2', 'UH1', '']
        assert res[2]['trace_ids'][0] == st2[2].id
        assert res[2]['trace_ids'][1] == st2[1].id
        assert res[2]['trace_ids'][2] == st2[0].id
        assert res[2]['trace_ids'][3] == st2[3].id
        assert res[2]['coincidence_sum'] == 1.4
        expected_keys = ['cft_peak_wmean', 'cft_std_wmean', 'cft_peaks',
                         'cft_stds']
        expected_types = [float, float, list, list]
        for item in res:
            for key, _type in zip(expected_keys, expected_types):
                assert key in item
                assert isinstance(item[key], _type)
        # check some of the detailed info
        ev = res[-1]
        assert round(abs(ev['cft_peak_wmean']-18.101139518271076), 5) == 0
        assert round(abs(ev['cft_std_wmean']-4.800051726246676), 5) == 0
        assert round(abs(ev['cft_peaks'][0]-18.985548683223936), 5) == 0
        assert round(abs(ev['cft_peaks'][1]-16.852175794415011), 5) == 0
        assert round(abs(ev['cft_peaks'][2]-18.64005853900883), 5) == 0
        assert round(abs(ev['cft_peaks'][3]-17.572363634564621), 5) == 0
        assert round(abs(ev['cft_stds'][0]-4.8909448258821362), 5) == 0
        assert round(abs(ev['cft_stds'][1]-4.4446373508521804), 5) == 0
        assert round(abs(ev['cft_stds'][2]-5.3499401252675964), 5) == 0
        assert round(abs(ev['cft_stds'][3]-4.2723814539487703), 5) == 0

    def test_coincidence_trigger_with_similarity_checking(self, testdata):
        """
        Test network coincidence trigger with cross correlation similarity
        checking of given event templates.
        """
        st = Stream()
        files = ["BW.UH1._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH2._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH3._.SHZ.D.2010.147.cut.slist.gz",
                 "BW.UH3._.SHN.D.2010.147.cut.slist.gz",
                 "BW.UH3._.SHE.D.2010.147.cut.slist.gz",
                 "BW.UH4._.EHZ.D.2010.147.cut.slist.gz"]
        for filename in files:
            st += read(testdata[filename])
        # some prefiltering used for UH network
        st.filter('bandpass', freqmin=10, freqmax=20)
        # set up template event streams
        times = ["2010-05-27T16:24:33.095000", "2010-05-27T16:27:30.370000"]
        templ = {}
        for t in times:
            t = UTCDateTime(t)
            st_ = st.select(station="UH3").slice(t, t + 2.5).copy()
            templ.setdefault("UH3", []).append(st_)
        times = ["2010-05-27T16:27:30.574999"]
        for t in times:
            t = UTCDateTime(t)
            st_ = st.select(station="UH1").slice(t, t + 2.5).copy()
            templ.setdefault("UH1", []).append(st_)
        # add another template with different SEED ID, it should be ignored
        # (this can happen when using many templates over a long time period
        # and instrument changes over time)
        st_ = st_.copy()
        for tr in st_:
            tr.stats.channel = 'X' + tr.stats.channel[1:]
        templ['UH1'].insert(0, st_)
        trace_ids = {"BW.UH1..SHZ": 1,
                     "BW.UH2..SHZ": 1,
                     "BW.UH3..SHZ": 1,
                     "BW.UH4..EHZ": 1}
        similarity_thresholds = {"UH1": 0.8, "UH3": 0.7}
        with warnings.catch_warnings(record=True) as w:
            # avoid getting influenced by the warning filters getting set up
            # differently in obspy-runtests.
            # (e.g. depending on options "-v" and "-q")
            warnings.resetwarnings()
            trig = coincidence_trigger(
                "classicstalta", 5, 1, st.copy(), 4, sta=0.5, lta=10,
                trace_ids=trace_ids, event_templates=templ,
                similarity_threshold=similarity_thresholds)
        # four warnings get raised
        assert len(w) == 4
        assert str(w[0].message) == \
            "At least one trace's ID was not found in the trace ID list and " \
            "was disregarded (BW.UH3..SHN)"
        assert str(w[1].message) == \
            "At least one trace's ID was not found in the trace ID list and " \
            "was disregarded (BW.UH3..SHE)"
        assert str(w[2].message) == \
            ('Skipping trace BW.UH1..XHZ in template correlation (not present '
             'in stream to check).')
        assert str(w[3].message) == \
            ("Skipping template(s) for station 'UH1': No common SEED IDs when "
             "comparing template (BW.UH1..XHZ) and data streams (BW.UH1..SHZ, "
             "BW.UH2..SHZ, BW.UH3..SHE, BW.UH3..SHN, BW.UH3..SHZ, "
             "BW.UH4..EHZ).")
        # check floats in resulting dictionary separately
        assert round(abs(trig[0].pop('duration')-3.96), 6) == 0
        assert round(abs(trig[1].pop('duration')-1.99), 6) == 0
        assert round(abs(trig[2].pop('duration')-1.92), 6) == 0
        assert round(abs(trig[3].pop('duration')-3.92), 6) == 0
        assert round(abs(trig[0]['similarity'].pop('UH1')-0.94149447384), 6) \
            == 0
        assert round(abs(trig[0]['similarity'].pop('UH3')-1), 6) == 0
        assert round(abs(trig[1]['similarity'].pop('UH1')-0.65228204570), 6) \
            == 0
        assert round(abs(trig[1]['similarity'].pop('UH3')-0.72679293429), 6) \
            == 0
        assert round(abs(trig[2]['similarity'].pop('UH1')-0.89404458774), 6) \
            == 0
        assert round(abs(trig[2]['similarity'].pop('UH3')-0.74581409371), 6) \
            == 0
        assert round(abs(trig[3]['similarity'].pop('UH1')-1), 6) == 0
        assert round(abs(trig[3]['similarity'].pop('UH3')-1), 6) == 0
        remaining_results = \
            [{'coincidence_sum': 4.0,
              'similarity': {},
              'stations': ['UH3', 'UH2', 'UH1', 'UH4'],
              'time': UTCDateTime(2010, 5, 27, 16, 24, 33, 210000),
              'trace_ids': ['BW.UH3..SHZ', 'BW.UH2..SHZ', 'BW.UH1..SHZ',
                            'BW.UH4..EHZ']},
             {'coincidence_sum': 3.0,
              'similarity': {},
              'stations': ['UH3', 'UH1', 'UH2'],
              'time': UTCDateTime(2010, 5, 27, 16, 25, 26, 710000),
              'trace_ids': ['BW.UH3..SHZ', 'BW.UH1..SHZ', 'BW.UH2..SHZ']},
             {'coincidence_sum': 3.0,
              'similarity': {},
              'stations': ['UH2', 'UH1', 'UH3'],
              'time': UTCDateTime(2010, 5, 27, 16, 27, 2, 260000),
              'trace_ids': ['BW.UH2..SHZ', 'BW.UH1..SHZ', 'BW.UH3..SHZ']},
             {'coincidence_sum': 4.0,
              'similarity': {},
              'stations': ['UH3', 'UH2', 'UH1', 'UH4'],
              'time': UTCDateTime(2010, 5, 27, 16, 27, 30, 510000),
              'trace_ids': ['BW.UH3..SHZ', 'BW.UH2..SHZ', 'BW.UH1..SHZ',
                            'BW.UH4..EHZ']}]
        assert trig == remaining_results

    def test_classic_sta_lta_c_python(self):
        """
        Test case for ctypes version of recursive_sta_lta
        """
        nsta, nlta = 5, 10
        c1 = classic_sta_lta(self.data, nsta, nlta)
        c2 = classic_sta_lta_py(self.data, nsta, nlta)
        assert np.allclose(c1, c2, rtol=1e-10)
        ref = np.array([0.38012302, 0.37704431, 0.47674533, 0.67992292])
        assert np.allclose(ref, c2[99:103])


class TestEnergyRatio():
    # parameterize ranges are based on chosen value of "a" with length 10
    @pytest.fixture(autouse=True, scope="function")
    def setup(self):
        self.a = np.empty(10)

    @pytest.mark.parametrize('nsta', (1, 2, 3, 4))
    def test_all_zero(self, nsta):
        self.a.fill(0)
        er = energy_ratio(self.a, nsta=nsta)
        assert_array_equal(er, 0)

    def test_arange(self):
        self.a = np.arange(10)
        er = energy_ratio(self.a, nsta=3)
        # Taken as the function output to keep track of regression bugs
        er_expected = [0., 0., 0., 10., 5.5, 3.793103, 2.98, 2.519481, 0., 0.]
        assert_array_almost_equal(er, er_expected)

    @pytest.mark.parametrize('nsta', (1, 2, 3, 4, 5))
    def test_all_ones(self, nsta):
        self.a.fill(1)
        # Forward and backward entries are symmetric -> expecting output '1'
        # Fill nsta on both sides with zero to return same length
        er = energy_ratio(self.a, nsta=nsta)
        er_exp = np.zeros_like(self.a)
        er_exp[nsta: len(self.a) - nsta + 1] = 1
        assert_array_equal(er, er_exp)

    @pytest.mark.parametrize('nsta', (6, 10, 20))
    def test_nsta_too_large(self, nsta):
        expected_msg = re.escape(
            f'nsta ({nsta}) must not be larger than half the length of '
            f'the data (10 samples).')
        with pytest.raises(ValueError, match=expected_msg):
            energy_ratio(self.a, nsta)

    @pytest.mark.parametrize('nsta', (0, -1, -10))
    def test_nsta_zero_or_less(self, nsta):
        expected_msg = re.escape(
            f'nsta ({nsta}) must not be equal to or less than zero.')
        with pytest.raises(ValueError, match=expected_msg):
            energy_ratio(self.a, nsta)


class TestModifiedEnergyRatio():
    # parameterize ranges are based on chosen value of "a" with length 10
    @pytest.fixture(autouse=True, scope="function")
    def setup(self):
        self.a = np.empty(10)

    @pytest.mark.parametrize('nsta', (1, 2, 3, 4))
    def test_all_zero(self, nsta):
        self.a.fill(0)
        er = modified_energy_ratio(self.a, nsta=nsta)
        assert_array_equal(er, 0)

    def test_arange(self):
        self.a = np.arange(10)
        er = modified_energy_ratio(self.a, nsta=3)
        # Taken as the function output to keep track of regression bugs
        er_expected = [0., 0., 0., 27000., 10648., 6821.722908, 5716.135872,
                       5485.637866, 0., 0.]
        assert_array_almost_equal(er, er_expected)

    @pytest.mark.parametrize('nsta', (1, 2, 3, 4, 5))
    def test_all_ones(self, nsta):
        self.a.fill(1)
        # Forward and backward entries are symmetric -> expecting output '1'
        # Fill nsta on both sides with zero to return same length
        er = modified_energy_ratio(self.a, nsta=nsta)
        er_exp = np.zeros_like(self.a)
        er_exp[nsta: len(self.a) - nsta + 1] = 1
        assert_array_equal(er, er_exp)

    @pytest.mark.parametrize('nsta', (6, 10, 20))
    def test_nsta_too_large(self, nsta):
        expected_msg = re.escape(
            f'nsta ({nsta}) must not be larger than half the length of '
            f'the data (10 samples).')
        with pytest.raises(ValueError, match=expected_msg):
            energy_ratio(self.a, nsta)

    @pytest.mark.parametrize('nsta', (0, -1, -10))
    def test_nsta_zero_or_less(self, nsta):
        expected_msg = re.escape(
            f'nsta ({nsta}) must not be equal to or less than zero.')
        with pytest.raises(ValueError, match=expected_msg):
            energy_ratio(self.a, nsta)



# -*- coding: utf-8 -*-
from .blockette import Blockette
from ..fields import Float, Integer, Loop, VariableString
from ..utils import format_resp


class Blockette058(Blockette):
    """
    Blockette 058: Channel Sensitivity/Gain Blockette.

    When used as a gain (stage ≠ 0), this blockette is the gain for this stage
    at the given frequency. Different stages may be at different frequencies.
    However, it is strongly recommended that the same frequency be used in all
    stages of a cascade, if possible. When used as a sensitivity(stage=0),
    this blockette is the sensitivity (in counts per ground motion) for the
    entire channel at a given frequency, and is also referred to as the
    overall gain. The frequency here may be different from the frequencies in
    the gain specifications, but should be the same if possible. If you use
    cascading (more than one filter stage), then SEED requires a gain for each
    stage. A final sensitivity (Blockette [58], stage = 0, is required. If you
    do not use cascading (only one stage), then SEED must see a gain, a
    sensitivity, or both.

    Sample:
    0580035 3 3.27680E+03 0.00000E+00 0
    """

    id = 58
    name = "Channel Sensitivity Gain"
    fields = [
        Integer(3, "Stage sequence number", 2),
        Float(4, "Sensitivity gain", 12, mask='%+1.5e'),
        Float(5, "Frequency", 12, mask='%+1.5e'),
        Integer(6, "Number of history values", 2),
        # REPEAT fields 7 — 9 for the Number of history values:
        Loop('History', "Number of history values", [
            Float(7, "Sensitivity for calibration", 12, mask='%+1.5e'),
            Float(8, "Frequency of calibration sensitivity", 12,
                  mask='%+1.5e'),
            VariableString(9, "Time of above calibration", 1, 22, 'T')
        ])
    ]

    def get_resp(self, station, channel, abbreviations):
        """
        Returns RESP string.
        """
        # This blockette can result in two different RESPs.
        blkt_type = self.stage_sequence_number
        if blkt_type != 0:
            string = \
                '#\t\t+                  +-------------------------------' + \
                '--------+                  +\n' + \
                '#\t\t+                  |       Channel Gain,' + \
                '%6s ch %s      |                  +\n' % (station, channel) +\
                '#\t\t+                  +-------------------------------' + \
                '--------+                  +\n'
        else:
            string = \
                '#\t\t+                  +--------------------------------' + \
                '-------+                  +\n' + \
                '#\t\t+                  |   Channel Sensitivity,' + \
                '%6s ch %s   |                  +\n' % (station, channel) + \
                '#\t\t+                  +--------------------------------' + \
                '-------+                  +\n'
        string += '#\t\t\n' + \
            'B058F03     Stage sequence number:                 %s\n' \
            % blkt_type
        if blkt_type != 0:
            string += \
                'B058F04     Gain:                                  %s\n' \
                % format_resp(self.sensitivity_gain, 6) + \
                'B058F05     Frequency of gain:                     %s HZ\n' \
                % format_resp(self.frequency, 6)
        else:
            string += \
                'B058F04     Sensitivity:                           %s\n' \
                % format_resp(self.sensitivity_gain, 6) + \
                'B058F05     Frequency of sensitivity:              %s HZ\n' \
                % format_resp(self.frequency, 6)
        string += \
            'B058F06     Number of calibrations:                %s\n' \
            % self.number_of_history_values
        if self.number_of_history_values > 1:
            string += \
                '#\t\tCalibrations:\n' + \
                '#\t\t i, sensitivity, frequency, time of calibration\n'
            for i in range(self.number_of_history_values):
                string += 'B058F07-08   %2s %13s %13s %s\n' % (
                    i,
                    format_resp(self.sensitivity_for_calibration[i], 6),
                    format_resp(
                        self.frequency_of_calibration_sensitivity[i], 6),
                    self.time_of_above_calibration[i].format_seed())
        elif self.number_of_history_values == 1:
            string += \
                '#\t\tCalibrations:\n' + \
                '#\t\t i, sensitivity, frequency, time of calibration\n' + \
                'B058F07-08    0 %13s %13s %s\n' \
                % (format_resp(self.sensitivity_for_calibration, 6),
                   format_resp(self.frequency_of_calibration_sensitivity, 6),
                   self.time_of_above_calibration.format_seed())
        string += '#\t\t\n'
        return string.encode()



# -*- coding: utf-8 -*-
"""
obspy.io.stationxml - StationXML read and write support for ObsPy
=================================================================

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""



# -*- coding: utf-8 -*-
"""
The obspy.realtime.signal test suite.
"""
import numpy as np
import pytest

from obspy import read
from obspy.core.stream import Stream
from obspy.realtime import RtTrace, signal


# some debug flags
NUM_PACKETS = 3


class TestRealTimeSignal():
    """
    The obspy.realtime.signal test suite.
    """
    @pytest.fixture(scope="function")
    def trace(self, testdata):
        # read test data as float64
        # this test data file is actually suffering from the floating point
        # issue handled in #3408, so set the buggy sampling rate to keep all
        # the tests in original form
        self.orig_trace = read(
            testdata['II.TLY.BHZ.SAC'], format="SAC", dtype=np.float64,
            round_sampling_interval=False)[0]
        # make really sure test data is float64
        self.orig_trace.data = np.require(self.orig_trace.data, np.float64)
        self.orig_trace_chunks = self.orig_trace / NUM_PACKETS
        trace = self.orig_trace.copy()
        yield trace
        # clear results
        self.filt_trace_data = None
        self.rt_trace = None
        self.rt_appended_traces = []

    def test_square(self, trace):
        """
        Testing np.square function.
        """
        # filtering manual
        self.filt_trace_data = np.square(trace)
        # filtering real time
        process_list = [(np.square, {})]
        self._run_rt_process(process_list)
        # check results
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_integrate(self, trace):
        """
        Testing integrate function.
        """
        # filtering manual
        self.filt_trace_data = signal.integrate(trace)
        # filtering real time
        process_list = [('integrate', {})]
        self._run_rt_process(process_list)
        # check results
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_differentiate(self, trace):
        """
        Testing differentiate function.
        """
        # filtering manual
        self.filt_trace_data = signal.differentiate(trace)
        # filtering real time
        process_list = [('differentiate', {})]
        self._run_rt_process(process_list)
        # check results
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_boxcar(self, trace):
        """
        Testing boxcar function.
        """
        options = {'width': 500}
        # filtering manual
        self.filt_trace_data = signal.boxcar(trace, **options)
        # filtering real time
        process_list = [('boxcar', options)]
        self._run_rt_process(process_list)
        # check results
        peak = np.amax(np.abs(self.rt_trace.data))
        assert round(abs(peak-566974.214), 3) == 0
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_scale(self, trace):
        """
        Testing scale function.
        """
        options = {'factor': 1000}
        # filtering manual
        self.filt_trace_data = signal.scale(trace, **options)
        # filtering real time
        process_list = [('scale', options)]
        self._run_rt_process(process_list)
        # check results
        peak = np.amax(np.abs(self.rt_trace.data))
        assert peak == 1045237000.0
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_offset(self, trace):
        """
        Testing offset function.
        """
        options = {'offset': 500}
        # filtering manual
        self.filt_trace_data = signal.offset(trace, **options)
        # filtering real time
        process_list = [('offset', options)]
        self._run_rt_process(process_list)
        # check results
        diff = self.rt_trace.data - self.orig_trace.data
        assert np.mean(diff) == 500
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_kurtosis(self, trace):
        """
        Testing kurtosis function.
        """
        options = {'win': 5}
        # filtering manual
        self.filt_trace_data = signal.kurtosis(trace, **options)
        # filtering real time
        process_list = [('kurtosis', options)]
        self._run_rt_process(process_list)
        # check results
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_abs(self, trace):
        """
        Testing np.abs function.
        """
        # filtering manual
        self.filt_trace_data = np.abs(trace)
        # filtering real time
        process_list = [(np.abs, {})]
        self._run_rt_process(process_list)
        # check results
        peak = np.amax(np.abs(self.rt_trace.data))
        assert peak == 1045237
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_tauc(self, trace):
        """
        Testing tauc function.
        """
        options = {'width': 60}
        # filtering manual
        self.filt_trace_data = signal.tauc(trace, **options)
        # filtering real time
        process_list = [('tauc', options)]
        self._run_rt_process(process_list)
        # check results
        peak = np.amax(np.abs(self.rt_trace.data))
        assert round(abs(peak-114.302), 3) == 0
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_mwp_integral(self, trace):
        """
        Testing mwpintegral functions.
        """
        options = {'mem_time': 240,
                   'ref_time': trace.stats.starttime + 301.506,
                   'max_time': 120,
                   'gain': 1.610210e+09}
        # filtering manual
        self.filt_trace_data = signal.mwpintegral(self.orig_trace.copy(),
                                                  **options)
        # filtering real time
        process_list = [('mwpintegral', options)]
        self._run_rt_process(process_list)
        # check results
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_mwp(self, trace):
        """
        Testing Mwp calculation using two processing functions.
        """
        epicentral_distance = 30.0855
        options = {'mem_time': 240,
                   'ref_time': trace.stats.starttime + 301.506,
                   'max_time': 120,
                   'gain': 1.610210e+09}
        # filtering manual
        trace.data = signal.integrate(trace)
        self.filt_trace_data = signal.mwpintegral(trace, **options)
        # filtering real time
        process_list = [('integrate', {}), ('mwpintegral', options)]
        self._run_rt_process(process_list)
        # check results
        peak = np.amax(np.abs(self.rt_trace.data))
        mwp = signal.calculate_mwp_mag(peak, epicentral_distance)
        assert round(abs(mwp-8.78902911791), 5) == 0
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)

    def test_combined(self, trace):
        """
        Testing combining integrate and differentiate functions.
        """
        # filtering manual
        trace.data = signal.integrate(trace)
        self.filt_trace_data = signal.differentiate(trace)
        # filtering real time
        process_list = [('int', {}), ('diff', {})]
        self._run_rt_process(process_list)
        # check results
        trace = self.orig_trace.copy()
        np.testing.assert_almost_equal(self.filt_trace_data,
                                       self.rt_trace.data)
        np.testing.assert_almost_equal(trace.data[1:], self.rt_trace.data[1:])
        np.testing.assert_almost_equal(trace.data[1:],
                                       self.filt_trace_data[1:])

    def _run_rt_process(self, process_list, max_length=None):
        """
        Helper function to create a RtTrace, register all given process
        functions and run the real time processing.
        """
        # assemble real time trace
        self.rt_trace = RtTrace(max_length=max_length)

        for (process, options) in process_list:
            self.rt_trace.register_rt_process(process, **options)

        # append packet data to RtTrace
        self.rt_appended_traces = []
        for trace in self.orig_trace_chunks:
            # process single trace
            result = self.rt_trace.append(trace, gap_overlap_check=True)
            # add to list of appended traces
            self.rt_appended_traces.append(result)

    def _plot_results(self):
        """
        Plots original, filtered original and real time processed traces into
        a single plot.
        """
        # plot only if test is started manually
        if __name__ != '__main__':
            return
        # create empty stream
        st = Stream()
        st.label = self._testMethodName
        # original trace
        self.orig_trace.label = "Original Trace"
        st += self.orig_trace
        # use header information of original trace with filtered trace data
        tr = self.orig_trace.copy()
        tr.data = self.filt_trace_data
        tr.label = "Filtered original Trace"
        st += tr
        # real processed chunks
        for i, tr in enumerate(self.rt_appended_traces):
            tr.label = "RT Chunk %02d" % (i + 1)
            st += tr
        # real time processed trace
        self.rt_trace.label = "RT Trace"
        st += self.rt_trace
        st.plot(automerge=False, color='blue', equal_scale=False)



# -*- coding: utf-8 -*-
"""
obspy.core.event.magnitude - The Magnitude class definition
===========================================================
This module provides a class hierarchy to consistently handle event metadata.
This class hierarchy is closely modelled after the de-facto standard format
`QuakeML <https://quake.ethz.ch/quakeml/>`_.

.. note::

    For handling additional information not covered by the QuakeML standard and
    how to output it to QuakeML see the :ref:`ObsPy Tutorial <quakeml-extra>`.

:copyright:
    The ObsPy Development Team (devs@obspy.org)
:license:
    GNU Lesser General Public License, Version 3
    (https://www.gnu.org/copyleft/lesser.html)
"""
from obspy import UTCDateTime
from obspy.core.event.base import (
    _event_type_class_factory, CreationInfo,
    WaveformStreamID, TimeWindow)
from obspy.core.event import ResourceIdentifier
from obspy.core.event.header import (
    AmplitudeCategory, AmplitudeUnit, EvaluationMode, EvaluationStatus,
    ATTRIBUTE_HAS_ERRORS)


__Magnitude = _event_type_class_factory(
    "__Magnitude",
    class_attributes=[("resource_id", ResourceIdentifier),
                      ("mag", float, ATTRIBUTE_HAS_ERRORS),
                      ("magnitude_type", str),
                      ("origin_id", ResourceIdentifier),
                      ("method_id", ResourceIdentifier),
                      ("station_count", int),
                      ("azimuthal_gap", float),
                      ("evaluation_mode", EvaluationMode),
                      ("evaluation_status", EvaluationStatus),
                      ("creation_info", CreationInfo)],
    class_contains=["comments", "station_magnitude_contributions"])


class Magnitude(__Magnitude):
    """
    Describes a magnitude which can, but does not need to be associated with an
    origin.

    Association with an origin is expressed with the optional attribute
    originID. It is either a combination of different magnitude estimations, or
    it represents the reported magnitude for the given event.

    :type resource_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param resource_id: Resource identifier of Magnitude.
    :type force_resource_id: bool, optional
    :param force_resource_id: If set to False, the automatic initialization of
        `resource_id` attribute in case it is not specified will be skipped.
    :type mag: float
    :param mag: Resulting magnitude value from combining values of type
        :class:`~obspy.core.event.magnitude.StationMagnitude`.
        If no estimations are available, this value can represent the
        reported magnitude.
    :type mag_errors: :class:`~obspy.core.event.base.QuantityError`
    :param mag_errors: AttribDict containing error quantities.
    :type magnitude_type: str, optional
    :param magnitude_type: Describes the type of magnitude. This is a free-text
        field because it is impossible to cover all existing magnitude type
        designations with an enumeration. Possible values are:

        * unspecified magnitude (``'M'``),
        * local magnitude (``'ML'``),
        * body wave magnitude (``'Mb'``),
        * surface wave magnitude (``'MS'``),
        * moment magnitude (``'Mw'``),
        * duration magnitude (``'Md'``)
        * coda magnitude (``'Mc'``)
        * ``'MH'``, ``'Mwp'``, ``'M50'``, ``'M100'``, etc.

    :type origin_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`,
        optional
    :param origin_id: Reference to an origin’s resource_id if the magnitude has
        an associated Origin.
    :type method_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`,
        optional
    :param method_id: Identifies the method of magnitude estimation. Users
        should avoid to give contradictory information in method_id and
        magnitude_type.
    :type station_count: int, optional
    :param station_count: Number of used stations for this magnitude
        computation.
    :type azimuthal_gap: float, optional
    :param azimuthal_gap: Azimuthal gap for this magnitude computation.
        Unit: deg
    :type evaluation_mode: str, optional
    :param evaluation_mode: Evaluation mode of Magnitude.
        See :class:`~obspy.core.event.header.EvaluationMode` for allowed
        values.
    :type evaluation_status: str, optional
    :param evaluation_status: Evaluation status of Magnitude.
        See :class:`~obspy.core.event.header.EvaluationStatus` for allowed
        values.
    :type comments: list of :class:`~obspy.core.event.base.Comment`, optional
    :param comments: Additional comments.
    :type station_magnitude_contributions: list of
        :class:`~obspy.core.event.magnitude.StationMagnitudeContribution`.
    :param station_magnitude_contributions: StationMagnitudeContribution
        instances associated with the Magnitude.
    :type creation_info: :class:`~obspy.core.event.base.CreationInfo`, optional
    :param creation_info: Creation information used to describe author,
        version, and creation time.

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__StationMagnitude = _event_type_class_factory(
    "__StationMagnitude",
    class_attributes=[("resource_id", ResourceIdentifier),
                      ("origin_id", ResourceIdentifier),
                      ("mag", float, ATTRIBUTE_HAS_ERRORS),
                      ("station_magnitude_type", str),
                      ("amplitude_id", ResourceIdentifier),
                      ("method_id", ResourceIdentifier),
                      ("waveform_id", WaveformStreamID),
                      ("creation_info", CreationInfo)],
    class_contains=["comments"])


class StationMagnitude(__StationMagnitude):
    """
    This class describes the magnitude derived from a single waveform stream.

    :type resource_id:
        :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param resource_id: Resource identifier of StationMagnitude.
    :type force_resource_id: bool, optional
    :param force_resource_id: If set to False, the automatic initialization of
        `resource_id` attribute in case it is not specified will be skipped.
    :type origin_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param origin_id: Reference to an origin’s ``resource_id`` if the
        StationMagnitude has an associated
        :class:`~obspy.core.event.origin.Origin`.
    :type mag: float
    :param mag: Estimated magnitude.
    :type mag_errors: :class:`~obspy.core.event.base.QuantityError`
    :param mag_errors: AttribDict containing error quantities.
    :type station_magnitude_type: str, optional
    :param station_magnitude_type: See
        :class:`~obspy.core.event.magnitude.Magnitude`
    :type amplitude_id:
        :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param amplitude_id: Identifies the data source of the StationMagnitude.
        For magnitudes derived from amplitudes in waveforms (e.g., local
        magnitude ML), amplitudeID points to publicID in class Amplitude.
    :type method_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param method_id: See :class:`~obspy.core.event.magnitude.Magnitude`
    :type waveform_id: :class:`~obspy.core.event.base.WaveformStreamID`,
        optional
    :param waveform_id: Identifies the waveform stream. This element can be
        helpful if no amplitude is referenced, or the amplitude is not
        available in the context. Otherwise, it would duplicate the waveform_id
        provided there and can be omitted.
    :type comments: list of :class:`~obspy.core.event.base.Comment`, optional
    :param comments: Additional comments.
    :type creation_info: :class:`~obspy.core.event.base.CreationInfo`, optional
    :param creation_info: Creation information used to describe author,
        version, and creation time.

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__StationMagnitudeContribution = _event_type_class_factory(
    "__StationMagnitudeContribution",
    class_attributes=[("station_magnitude_id", ResourceIdentifier),
                      ("residual", float),
                      ("weight", float)])


class StationMagnitudeContribution(__StationMagnitudeContribution):
    """
    This class describes the weighting of magnitude values from several
    StationMagnitude objects for computing a network magnitude estimation.

    :type station_magnitude_id:
        :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param station_magnitude_id: Refers to the resource_id of a
        StationMagnitude object.
    :type residual: float, optional
    :param residual: Residual of magnitude computation.
    :type weight: float, optional
    :param weight: Weight of the magnitude value from class StationMagnitude
        for computing the magnitude value in class Magnitude. Note that there
        is no rule for the sum of the weights of all station magnitude
        contributions to a specific network magnitude. In particular, the
        weights are not required to sum up to unity.

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


__Amplitude = _event_type_class_factory(
    "__Amplitude",
    class_attributes=[("resource_id", ResourceIdentifier),
                      ("generic_amplitude", float, ATTRIBUTE_HAS_ERRORS),
                      ("type", str),
                      ("category", AmplitudeCategory),
                      ("unit", AmplitudeUnit),
                      ("method_id", ResourceIdentifier),
                      ("period", float, ATTRIBUTE_HAS_ERRORS),
                      ("snr", float),
                      ("time_window", TimeWindow),
                      ("pick_id", ResourceIdentifier),
                      ("waveform_id", WaveformStreamID),
                      ("filter_id", ResourceIdentifier),
                      ("scaling_time", UTCDateTime, ATTRIBUTE_HAS_ERRORS),
                      ("magnitude_hint", str),
                      ("evaluation_mode", EvaluationMode),
                      ("evaluation_status", EvaluationStatus),
                      ("creation_info", CreationInfo)],
    class_contains=["comments"])


class Amplitude(__Amplitude):
    """
    This class represents a quantification of the waveform anomaly, usually a
    single amplitude measurement or a measurement of the visible signal
    duration for duration magnitudes.

    :type resource_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param resource_id: Resource identifier of Amplitude.
    :type force_resource_id: bool, optional
    :param force_resource_id: If set to False, the automatic initialization of
        `resource_id` attribute in case it is not specified will be skipped.
    :type generic_amplitude: float
    :param generic_amplitude: Measured amplitude value for the given
        waveformID. Note that this attribute can describe different physical
        quantities, depending on the type and category of the amplitude. These
        can be, e.g., displacement, velocity, or a period. If the only
        amplitude information is a period, it has to specified here, not in the
        period attribute. The latter can be used if the amplitude measurement
        contains information on, e.g., displacement and an additional period.
        Since the physical quantity described by this attribute is not fixed,
        the unit of measurement cannot be defined in advance. However, the
        quantity has to be specified in SI base units. The enumeration given in
        attribute unit provides the most likely units that could be needed
        here. For clarity, using the optional unit attribute is highly
        encouraged.
    :type generic_amplitude_errors:
        :class:`~obspy.core.event.base.QuantityError`
    :param generic_amplitude_errors: AttribDict containing error quantities.
    :type type: str, optional
    :param type: Describes the type of amplitude using the nomenclature from
        Storchak et al. (2003). Possible values are:

        * unspecified amplitude reading (``'A'``),
        * amplitude reading for local magnitude (``'AML'``),
        * amplitude reading for body wave magnitude (``'AMB'``),
        * amplitude reading for surface wave magnitude (``'AMS'``), and
        * time of visible end of record for duration magnitude (``'END'``).

    :type category: str, optional
    :param category:  Amplitude category.  This attribute describes the way the
        waveform trace is evaluated to derive an amplitude value. This can be
        just reading a single value for a given point in time (point), taking a
        mean value over a time interval (mean), integrating the trace over a
        time interval (integral), specifying just a time interval (duration),
        or evaluating a period (period).
        See :class:`~obspy.core.event.header.AmplitudeCategory` for allowed
        values.
    :type unit: str, optional
    :param unit: Amplitude unit. This attribute provides the most likely
        measurement units for the physical quantity described in the
        genericAmplitude attribute. Possible values are specified as
        combinations of SI base units.
        See :class:`~obspy.core.event.header.AmplitudeUnit` for allowed
        values.
    :type method_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param method_id: Describes the method of amplitude determination.
    :type period: float, optional
    :param period: Dominant period in the timeWindow in case of amplitude
        measurements. Not used for duration magnitude.  Unit: s
    :type snr: float, optional
    :param snr: Signal-to-noise ratio of the spectrogram at the location the
        amplitude was measured.
    :type time_window: :class:`~obspy.core.event.base.TimeWindow`, optional
    :param time_window: Description of the time window used for amplitude
        measurement. Recommended for duration magnitudes.
    :type pick_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param pick_id: Refers to the ``resource_id`` of an associated
        :class:`~obspy.core.event.origin.Pick` object.
    :type waveform_id: :class:`~obspy.core.event.base.WaveformStreamID`,
    :param waveform_id: Identifies the waveform stream on which the amplitude
        was measured.
    :type filter_id: :class:`~obspy.core.event.resourceid.ResourceIdentifier`
    :param filter_id: Identifies the filter or filter setup used for filtering
        the waveform stream referenced by ``waveform_id``.
    :type scaling_time: :class:`~obspy.core.utcdatetime.UTCDateTime`, optional
    :param scaling_time: Scaling time for amplitude measurement.
    :type scaling_time_errors: :class:`~obspy.core.event.base.QuantityError`
    :param scaling_time_errors: AttribDict containing error quantities.
    :type magnitude_hint: str, optional
    :param magnitude_hint: Type of magnitude the amplitude measurement is used
        for.  This is a free-text field because it is impossible to cover all
        existing magnitude type designations with an enumeration. Possible
        values are:

        * unspecified magnitude (``'M'``),
        * local magnitude (``'ML'``),
        * body wave magnitude (``'Mb'``),
        * surface wave magnitude (``'MS'``),
        * moment magnitude (``'Mw'``),
        * duration magnitude (``'Md'``)
        * coda magnitude (``'Mc'``)
        * ``'MH'``, ``'Mwp'``, ``'M50'``, ``'M100'``, etc.

    :type evaluation_mode: str, optional
    :param evaluation_mode: Evaluation mode of Amplitude.
        See :class:`~obspy.core.event.header.EvaluationMode` for allowed
        values.
    :type evaluation_status: str, optional
    :param evaluation_status: Evaluation status of Amplitude.
        See :class:`~obspy.core.event.header.EvaluationStatus` for allowed
        values.
    :type comments: list of :class:`~obspy.core.event.base.Comment`, optional
    :param comments: Additional comments.
    :type creation_info: :class:`~obspy.core.event.base.CreationInfo`, optional
    :param creation_info: CreationInfo for the Amplitude object.

    .. note::

        For handling additional information not covered by the QuakeML
        standard and how to output it to QuakeML see the
        :ref:`ObsPy Tutorial <quakeml-extra>`.
    """


if __name__ == '__main__':
    import doctest
    doctest.testmod(exclude_empty=True)



# -*- coding: utf-8 -*-
"""
ALSEP module helper functions and data.
"""
from obspy import UTCDateTime


def interp(x1, x2, x3, x4):
    """Interpolate by Lagrange polynomial with two points before and after"""
    return int(((x2 + x3) * 4 - (x1 + x4) + 3) / 6)


def get_utc(year, msec_of_year):
    """Calculate date and time in UTC from year and millisecond of year"""
    return UTCDateTime(year, month=1, day=1) + msec_of_year / 1000.0


def doy2utc(year, doy):
    """Calculate date and time in UTC from year and day of year"""
    return UTCDateTime(year, julday=doy)


def check_date(apollo_station, utc):
    """Check observation date"""
    if apollo_station == 11:
        # 1969/202 - 1969/214, 1969/231 - 1969/237
        if doy2utc(1969, 202) <= utc < doy2utc(1969, 215):
            return True
        if doy2utc(1969, 231) <= utc < doy2utc(1969, 238):
            return True
    elif apollo_station == 12:
        # 1969/323-1977/273
        if doy2utc(1969, 323) <= utc < doy2utc(1977, 274):
            return True
    elif apollo_station == 14:
        # 1971/036-1977/273
        if doy2utc(1971, 36) <= utc < doy2utc(1977, 274):
            return True
    elif apollo_station == 15:
        # 1971/212-1977/273
        if doy2utc(1971, 212) <= utc < doy2utc(1977, 274):
            return True
    elif apollo_station == 16:
        # 1972/112-1977/273
        if doy2utc(1972, 112) <= utc < doy2utc(1977, 274):
            return True
    elif apollo_station == 17:
        # LSPE: 1976/228-1977/115
        # LSG: 1976/061-1977/273
        if doy2utc(1976, 61) <= utc < doy2utc(1977, 273):
            return True
    return False


def check_sync_code(sync_code):
    """Check sync pattern barker code"""
    return True if sync_code == 0b11100010010 else False
